from rest import decorators as rd
from rest import views as rv
from rest import models as rm
from account import models as am
import requests

@rd.url(r'^email/bounced/$')
@rd.url(r'^email/bounced/(?P<pk>\d+)$')
@rd.login_required
def handleBounced(request, pk=None):
    return am.BounceHistory.on_rest_request(request, pk)


@rd.url(r'^notifications/$')
@rd.url(r'^notifications/(?P<pk>\d+)$')
@rd.perm_required("manage_users")
def handleNotifications(request, pk=None):
    return am.NotificationRecord.on_rest_request(request, pk)


# BEGIN AWS SNS HANDLER

EMAIL_MODEL_CACHE = {}

def handleBounce(request):
    msg = request.DATA.asUberDict()
    if not msg.bounce or not isinstance(msg.bounce.bouncedRecipients, list):
        return rv.restStatus(request, True)

    for who in msg.bounce.bouncedRecipients:
        # kind, address, reason, reporter=None, code=None, source=None, source_ip=None, user=None
        am.BounceHistory.log(
            kind="email",
            address=who.emailAddress,
            reason=who.diagnosticCode,
            reporter=msg.bounce.reportingMTA,
            code=who.status,
            source=msg.mail.source,
            source_ip=msg.mail.sourceIp)
    return rv.restStatus(request, True)

def getEmailHandler(name):
    if name in EMAIL_MODEL_CACHE:
        return EMAIL_MODEL_CACHE[name]
    a_name, m_name = name.split(".")
    model = rm.RestModel.getModel(a_name, m_name)
    EMAIL_MODEL_CACHE[name] = model
    return model

def onValidEmail(msg):
    handlers = EMAIL_MODEL_HANDLERS.get(msg.to.lower(), None)
    if not handlers:
        return False
    if isinstance(handlers, list):
        for name in handlers:
            handler = getEmailHandler(name)
            if handler and hasattr(handler, "on_email"):
                handler.on_email(msg)
    else:
        handler = getEmailHandler(handlers)
        if handler and hasattr(handler, "on_email"):
            handler.on_email(msg)

@rd.url(r'^aws/sns$')
@rd.url(r'^aws/sns/$')
def email_handler(request):
    if not request.is_sns:
        print((request.META))
        print((request.body))
        return rv.restStatus(request, False, error="invalid type")

    if request.sns_type == "Bounce":
        return handleBounce(request)

    if request.sns_type == "SubscriptionConfirmation":
        url = request.DATA.get("SubscribeURL")
        res = requests.get(url)
        return rv.restStatus(request, True)

    if request.sns_type == "email":
        source = request.DATA.get("mail.source")
        # this is a list
        destination = request.DATA.get("mail.destination")
        subject = request.DATA.get("mail.commonHeaders.subject")
        content = request.DATA.get("content")
        if content:
            try:
                msg = mailman.parse_raw_message(content)
                # print("begin email")
                # print(msg)
                # print("end email")
                onValidEmail(UberDict.fromdict(msg))
                # parseTicketEmail(msg)
            except Exception as err:
                stack = str(traceback.format_exc())
                Member.notifyWithPermission("rest_errors", "error: {}\n<br>{}\n<br>\n{}".format(str(err), stack, content), context={}, email_only=True)
    print(("sns_type = {}".format(request.sns_type)))
    return rv.restStatus(request, True)




# Generated by Django 4.0.2 on 2022-02-27 04:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0003_member_phone_number'),
    ]

    operations = [
        migrations.AddField(
            model_name='group',
            name='modified',
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name='group',
            name='created',
            field=models.DateTimeField(auto_now_add=True),
        ),
    ]



from rest import helpers
import json
import time
import datetime
import math
from decimal import Decimal
import pprint


def serialize(data):
    return toJSON(data)


def toJSON(data, **kwargs):
    return json.dumps(data, cls=JSONEncoderExt)
    # return json.dumps(data, **kwargs)


def prettyJSON(data):
    return json.dumps(data, cls=JSONEncoderExt, sort_keys=True, indent=4, separators=(',', ': '))


def prettyJSON2(data):
    return pprint.pformat(data)


try:
    import ujson
    def toJSON(data, **kwargs):
        # helpers.log_print(data)
        return ujson.dumps(data)
except Exception:
    helpers.log_print("recommend installing ujson!")


def toString(value):
    if isinstance(value, bytes):
        value = value.decode()
    elif isinstance(value, bytearray):
        value = value.decode("utf-8")
    elif isinstance(value, (int, float)):
        value = str(value)
    return value


# is this useless??? we have rest_serialize
class JSONEncoderExt(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return time.mktime(obj.timetuple())
        elif isinstance(obj, datetime.date):
            return obj.strftime("%Y/%m/%d")
        elif isinstance(obj, Decimal):
            if obj.is_nan():
                return 0.0
            return float(obj)
        elif isinstance(obj, float):
            if math.isnan(obj):
                return 0.0
        elif isinstance(obj, set):
            # helpers.log_error(obj)
            return str(obj)
        elif isinstance(obj, (bytes, bytearray)):
            return toString(obj)
        try:
            return super().default(obj)
        except Exception:
            pass
        return "not parsable"




# Generated by Django 4.1.3 on 2022-12-16 18:03

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('incident', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='event',
            name='component',
            field=models.SlugField(blank=True, default=None, max_length=250, null=True),
        ),
        migrations.AddField(
            model_name='event',
            name='component_id',
            field=models.IntegerField(blank=True, default=None, null=True),
        ),
    ]




from django.db import models
from rest import models as rm
from rest import fields as rf
from rest import settings
from objict import objict
from datetime import datetime


class UserPassKey(models.Model, rm.RestModel):
    """
    MemberDevice Model tracks personal devices associated with a user.
    This can include mobile and desktop devices.
    """
    class RestMeta:
        VIEW_PERMS = ["manage_users", "owner"]
        CAN_DELETE = True
        VIEW_PERMS_MEMBER_FIELD = "member"
        GRAPHS = {
            "default": {
                "fields": [
                    "id", "created", "modified", "last_used",
                    "uuid", "rp_id", "is_enabled", 
                    "name", "platform", "info"
                ],
            }
        }
        
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    member = models.ForeignKey("account.Member", related_name="passkeys", on_delete=models.CASCADE)

    uuid = models.CharField(max_length=255, unique=True)
    rp_id = models.CharField(max_length=255, null=False, db_index=True)
    is_enabled = models.BooleanField(default=True, db_index=True)

    name = models.CharField(max_length=255)
    platform = models.CharField(max_length=255, default='')
    last_used = models.DateTimeField(null=True, default=None)
    token = models.CharField(max_length=255, null=False)

    info = rf.JSONField()

    def touch(self):
        self.last_used = datetime.now()
        self.save()






# Generated by Django 4.0.2 on 2022-02-22 01:23

from django.db import migrations, models
import django.db.models.deletion
import rest.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('medialib', '__first__'),
        ('account', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Product',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('oid', models.CharField(blank=True, db_index=True, default='', max_length=250)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('archived', models.BooleanField(blank=True, default=False)),
                ('name', models.CharField(max_length=200)),
                ('description', models.TextField(blank=True, default=None, null=True)),
                ('kind', models.CharField(blank=True, default=None, max_length=64, null=True)),
                ('is_public', models.BooleanField(blank=True, default=False)),
            ],
            bases=(models.Model, rest.models.RestModel, rest.models.MetaDataModel),
        ),
        migrations.CreateModel(
            name='Release',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('notes', models.TextField(blank=True, default=None, null=True)),
                ('version_num', models.IntegerField(db_index=True, default=1)),
                ('version_str', models.CharField(blank=True, max_length=32, null=True)),
                ('media', models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.PROTECT, to='medialib.mediaitem')),
                ('owner', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='account.member')),
                ('product', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='releases', to='pushit.product')),
            ],
            bases=(models.Model, rest.models.RestModel, rest.models.MetaDataModel),
        ),
        migrations.CreateModel(
            name='ProductMetaData',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('category', models.CharField(blank=True, db_index=True, default=None, max_length=32, null=True)),
                ('key', models.CharField(db_index=True, max_length=80)),
                ('value_format', models.CharField(max_length=16)),
                ('value', models.TextField()),
                ('int_value', models.IntegerField(blank=True, default=None, null=True)),
                ('float_value', models.IntegerField(blank=True, default=None, null=True)),
                ('parent', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='properties', to='pushit.product')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='product',
            name='beta',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='pushit.release'),
        ),
        migrations.AddField(
            model_name='product',
            name='current',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='pushit.release'),
        ),
        migrations.AddField(
            model_name='product',
            name='group',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='account.group'),
        ),
        migrations.AddField(
            model_name='product',
            name='library',
            field=models.ForeignKey(blank=True, default=None, null=True, on_delete=django.db.models.deletion.CASCADE, to='medialib.medialibrary'),
        ),
        migrations.AddField(
            model_name='product',
            name='owner',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='account.member'),
        ),
    ]



from __future__ import unicode_literals
from django.core.management.base import BaseCommand, CommandError
from redis_metrics.utils import get_r


class Command(BaseCommand):
    args = '<gauge-key>'
    help = "Removes a Gauge and its data"

    def handle(self, *args, **options):
        if not len(args) == 1:
            raise CommandError("You must provide a gauge name")
        r = get_r()
        r.delete_gauge(args[0])



# Copyright (C) 2010 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Client for discovery based APIs.

A client library for Google's discovery based APIs.
"""

__author__ = 'jcgregorio@google.com (Joe Gregorio)'
__all__ = [
    'build',
    'build_from_document',
    'fix_method_name',
    'key2param',
    ]


# Standard library imports
import copy
from email.mime.multipart import MIMEMultipart
from email.mime.nonmultipart import MIMENonMultipart
import keyword
import logging
import mimetypes
import os
import re
import urllib.request, urllib.parse, urllib.error
import urllib.parse

try:
  from urllib.parse import parse_qsl
except ImportError:
  from cgi import parse_qsl

# Third-party imports
from medialib.youtube import httplib2
from . import mimeparse
from medialib.youtube import uritemplate

# Local imports
from .errors import HttpError
from .errors import InvalidJsonError
from .errors import MediaUploadSizeError
from .errors import UnacceptableMimeTypeError
from .errors import UnknownApiNameOrVersion
from .errors import UnknownFileType
from .http import HttpRequest
from .http import MediaFileUpload
from .http import MediaUpload
from .model import JsonModel
from .model import MediaModel
from .model import RawModel
from .schema import Schemas
from medialib.youtube.oauth2client.anyjson import simplejson
from medialib.youtube.oauth2client.util import _add_query_parameter
from medialib.youtube.oauth2client.util import positional


# The client library requires a version of httplib2 that supports RETRIES.
httplib2.RETRIES = 1

logger = logging.getLogger(__name__)

URITEMPLATE = re.compile('{[^}]*}')
VARNAME = re.compile('[a-zA-Z0-9_-]+')
DISCOVERY_URI = ('https://www.googleapis.com/discovery/v1/apis/'
                 '{api}/{apiVersion}/rest')
DEFAULT_METHOD_DOC = 'A description of how to use this function'
HTTP_PAYLOAD_METHODS = frozenset(['PUT', 'POST', 'PATCH'])
_MEDIA_SIZE_BIT_SHIFTS = {'KB': 10, 'MB': 20, 'GB': 30, 'TB': 40}
BODY_PARAMETER_DEFAULT_VALUE = {
    'description': 'The request body.',
    'type': 'object',
    'required': True,
}
MEDIA_BODY_PARAMETER_DEFAULT_VALUE = {
    'description': ('The filename of the media request body, or an instance '
                    'of a MediaUpload object.'),
    'type': 'string',
    'required': False,
}

# Parameters accepted by the stack, but not visible via discovery.
# TODO(dhermes): Remove 'userip' in 'v2'.
STACK_QUERY_PARAMETERS = frozenset(['trace', 'pp', 'userip', 'strict'])
STACK_QUERY_PARAMETER_DEFAULT_VALUE = {'type': 'string', 'location': 'query'}

# Library-specific reserved words beyond Python keywords.
RESERVED_WORDS = frozenset(['body'])


def fix_method_name(name):
  """Fix method names to avoid reserved word conflicts.

  Args:
    name: string, method name.

  Returns:
    The name with a '_' prefixed if the name is a reserved word.
  """
  if keyword.iskeyword(name) or name in RESERVED_WORDS:
    return name + '_'
  else:
    return name


def key2param(key):
  """Converts key names into parameter names.

  For example, converting "max-results" -> "max_results"

  Args:
    key: string, the method key name.

  Returns:
    A safe method name based on the key name.
  """
  result = []
  key = list(key)
  if not key[0].isalpha():
    result.append('x')
  for c in key:
    if c.isalnum():
      result.append(c)
    else:
      result.append('_')

  return ''.join(result)


@positional(2)
def build(serviceName,
          version,
          http=None,
          discoveryServiceUrl=DISCOVERY_URI,
          developerKey=None,
          model=None,
          requestBuilder=HttpRequest):
  """Construct a Resource for interacting with an API.

  Construct a Resource object for interacting with an API. The serviceName and
  version are the names from the Discovery service.

  Args:
    serviceName: string, name of the service.
    version: string, the version of the service.
    http: httplib2.Http, An instance of httplib2.Http or something that acts
      like it that HTTP requests will be made through.
    discoveryServiceUrl: string, a URI Template that points to the location of
      the discovery service. It should have two parameters {api} and
      {apiVersion} that when filled in produce an absolute URI to the discovery
      document for that service.
    developerKey: string, key obtained from
      https://code.google.com/apis/console.
    model: apiclient.Model, converts to and from the wire format.
    requestBuilder: apiclient.http.HttpRequest, encapsulator for an HTTP
      request.

  Returns:
    A Resource object with methods for interacting with the service.
  """
  params = {
      'api': serviceName,
      'apiVersion': version
      }

  if http is None:
    http = httplib2.Http()

  requested_url = uritemplate.expand(discoveryServiceUrl, params)

  # REMOTE_ADDR is defined by the CGI spec [RFC3875] as the environment
  # variable that contains the network address of the client sending the
  # request. If it exists then add that to the request for the discovery
  # document to avoid exceeding the quota on discovery requests.
  if 'REMOTE_ADDR' in os.environ:
    requested_url = _add_query_parameter(requested_url, 'userIp',
                                         os.environ['REMOTE_ADDR'])
  logger.info('URL being requested: %s' % requested_url)

  resp, content = http.request(requested_url)

  if resp.status == 404:
    raise UnknownApiNameOrVersion("name: %s  version: %s" % (serviceName,
                                                            version))
  if resp.status >= 400:
    raise HttpError(resp, content, uri=requested_url)

  try:
    service = simplejson.loads(content)
  except ValueError as e:
    logger.error('Failed to parse as JSON: ' + content)
    raise InvalidJsonError()

  return build_from_document(content, base=discoveryServiceUrl, http=http,
      developerKey=developerKey, model=model, requestBuilder=requestBuilder)


@positional(1)
def build_from_document(
    service,
    base=None,
    future=None,
    http=None,
    developerKey=None,
    model=None,
    requestBuilder=HttpRequest):
  """Create a Resource for interacting with an API.

  Same as `build()`, but constructs the Resource object from a discovery
  document that is it given, as opposed to retrieving one over HTTP.

  Args:
    service: string or object, the JSON discovery document describing the API.
      The value passed in may either be the JSON string or the deserialized
      JSON.
    base: string, base URI for all HTTP requests, usually the discovery URI.
      This parameter is no longer used as rootUrl and servicePath are included
      within the discovery document. (deprecated)
    future: string, discovery document with future capabilities (deprecated).
    http: httplib2.Http, An instance of httplib2.Http or something that acts
      like it that HTTP requests will be made through.
    developerKey: string, Key for controlling API usage, generated
      from the API Console.
    model: Model class instance that serializes and de-serializes requests and
      responses.
    requestBuilder: Takes an http request and packages it up to be executed.

  Returns:
    A Resource object with methods for interacting with the service.
  """

  # future is no longer used.
  future = {}

  if isinstance(service, str):
    service = simplejson.loads(service)
  base = urllib.parse.urljoin(service['rootUrl'], service['servicePath'])
  schema = Schemas(service)

  if model is None:
    features = service.get('features', [])
    model = JsonModel('dataWrapper' in features)
  return Resource(http=http, baseUrl=base, model=model,
                  developerKey=developerKey, requestBuilder=requestBuilder,
                  resourceDesc=service, rootDesc=service, schema=schema)


def _cast(value, schema_type):
  """Convert value to a string based on JSON Schema type.

  See http://tools.ietf.org/html/draft-zyp-json-schema-03 for more details on
  JSON Schema.

  Args:
    value: any, the value to convert
    schema_type: string, the type that value should be interpreted as

  Returns:
    A string representation of 'value' based on the schema_type.
  """
  if schema_type == 'string':
    if type(value) == type('') or type(value) == type(''):
      return value
    else:
      return str(value)
  elif schema_type == 'integer':
    return str(int(value))
  elif schema_type == 'number':
    return str(float(value))
  elif schema_type == 'boolean':
    return str(bool(value)).lower()
  else:
    if type(value) == type('') or type(value) == type(''):
      return value
    else:
      return str(value)


def _media_size_to_long(maxSize):
  """Convert a string media size, such as 10GB or 3TB into an integer.

  Args:
    maxSize: string, size as a string, such as 2MB or 7GB.

  Returns:
    The size as an integer value.
  """
  if len(maxSize) < 2:
    return 0
  units = maxSize[-2:].upper()
  bit_shift = _MEDIA_SIZE_BIT_SHIFTS.get(units)
  if bit_shift is not None:
    return int(maxSize[:-2]) << bit_shift
  else:
    return int(maxSize)


def _media_path_url_from_info(root_desc, path_url):
  """Creates an absolute media path URL.

  Constructed using the API root URI and service path from the discovery
  document and the relative path for the API method.

  Args:
    root_desc: Dictionary; the entire original deserialized discovery document.
    path_url: String; the relative URL for the API method. Relative to the API
        root, which is specified in the discovery document.

  Returns:
    String; the absolute URI for media upload for the API method.
  """
  return '%(root)supload/%(service_path)s%(path)s' % {
      'root': root_desc['rootUrl'],
      'service_path': root_desc['servicePath'],
      'path': path_url,
  }


def _fix_up_parameters(method_desc, root_desc, http_method):
  """Updates parameters of an API method with values specific to this library.

  Specifically, adds whatever global parameters are specified by the API to the
  parameters for the individual method. Also adds parameters which don't
  appear in the discovery document, but are available to all discovery based
  APIs (these are listed in STACK_QUERY_PARAMETERS).

  SIDE EFFECTS: This updates the parameters dictionary object in the method
  description.

  Args:
    method_desc: Dictionary with metadata describing an API method. Value comes
        from the dictionary of methods stored in the 'methods' key in the
        deserialized discovery document.
    root_desc: Dictionary; the entire original deserialized discovery document.
    http_method: String; the HTTP method used to call the API method described
        in method_desc.

  Returns:
    The updated Dictionary stored in the 'parameters' key of the method
        description dictionary.
  """
  parameters = method_desc.setdefault('parameters', {})

  # Add in the parameters common to all methods.
  for name, description in list(root_desc.get('parameters', {}).items()):
    parameters[name] = description

  # Add in undocumented query parameters.
  for name in STACK_QUERY_PARAMETERS:
    parameters[name] = STACK_QUERY_PARAMETER_DEFAULT_VALUE.copy()

  # Add 'body' (our own reserved word) to parameters if the method supports
  # a request payload.
  if http_method in HTTP_PAYLOAD_METHODS and 'request' in method_desc:
    body = BODY_PARAMETER_DEFAULT_VALUE.copy()
    body.update(method_desc['request'])
    parameters['body'] = body

  return parameters


def _fix_up_media_upload(method_desc, root_desc, path_url, parameters):
  """Updates parameters of API by adding 'media_body' if supported by method.

  SIDE EFFECTS: If the method supports media upload and has a required body,
  sets body to be optional (required=False) instead. Also, if there is a
  'mediaUpload' in the method description, adds 'media_upload' key to
  parameters.

  Args:
    method_desc: Dictionary with metadata describing an API method. Value comes
        from the dictionary of methods stored in the 'methods' key in the
        deserialized discovery document.
    root_desc: Dictionary; the entire original deserialized discovery document.
    path_url: String; the relative URL for the API method. Relative to the API
        root, which is specified in the discovery document.
    parameters: A dictionary describing method parameters for method described
        in method_desc.

  Returns:
    Triple (accept, max_size, media_path_url) where:
      - accept is a list of strings representing what content types are
        accepted for media upload. Defaults to empty list if not in the
        discovery document.
      - max_size is a long representing the max size in bytes allowed for a
        media upload. Defaults to 0L if not in the discovery document.
      - media_path_url is a String; the absolute URI for media upload for the
        API method. Constructed using the API root URI and service path from
        the discovery document and the relative path for the API method. If
        media upload is not supported, this is None.
  """
  media_upload = method_desc.get('mediaUpload', {})
  accept = media_upload.get('accept', [])
  max_size = _media_size_to_long(media_upload.get('maxSize', ''))
  media_path_url = None

  if media_upload:
    media_path_url = _media_path_url_from_info(root_desc, path_url)
    parameters['media_body'] = MEDIA_BODY_PARAMETER_DEFAULT_VALUE.copy()
    if 'body' in parameters:
      parameters['body']['required'] = False

  return accept, max_size, media_path_url


def _fix_up_method_description(method_desc, root_desc):
  """Updates a method description in a discovery document.

  SIDE EFFECTS: Changes the parameters dictionary in the method description with
  extra parameters which are used locally.

  Args:
    method_desc: Dictionary with metadata describing an API method. Value comes
        from the dictionary of methods stored in the 'methods' key in the
        deserialized discovery document.
    root_desc: Dictionary; the entire original deserialized discovery document.

  Returns:
    Tuple (path_url, http_method, method_id, accept, max_size, media_path_url)
    where:
      - path_url is a String; the relative URL for the API method. Relative to
        the API root, which is specified in the discovery document.
      - http_method is a String; the HTTP method used to call the API method
        described in the method description.
      - method_id is a String; the name of the RPC method associated with the
        API method, and is in the method description in the 'id' key.
      - accept is a list of strings representing what content types are
        accepted for media upload. Defaults to empty list if not in the
        discovery document.
      - max_size is a long representing the max size in bytes allowed for a
        media upload. Defaults to 0L if not in the discovery document.
      - media_path_url is a String; the absolute URI for media upload for the
        API method. Constructed using the API root URI and service path from
        the discovery document and the relative path for the API method. If
        media upload is not supported, this is None.
  """
  path_url = method_desc['path']
  http_method = method_desc['httpMethod']
  method_id = method_desc['id']

  parameters = _fix_up_parameters(method_desc, root_desc, http_method)
  # Order is important. `_fix_up_media_upload` needs `method_desc` to have a
  # 'parameters' key and needs to know if there is a 'body' parameter because it
  # also sets a 'media_body' parameter.
  accept, max_size, media_path_url = _fix_up_media_upload(
      method_desc, root_desc, path_url, parameters)

  return path_url, http_method, method_id, accept, max_size, media_path_url


# TODO(dhermes): Convert this class to ResourceMethod and make it callable
class ResourceMethodParameters(object):
  """Represents the parameters associated with a method.

  Attributes:
    argmap: Map from method parameter name (string) to query parameter name
        (string).
    required_params: List of required parameters (represented by parameter
        name as string).
    repeated_params: List of repeated parameters (represented by parameter
        name as string).
    pattern_params: Map from method parameter name (string) to regular
        expression (as a string). If the pattern is set for a parameter, the
        value for that parameter must match the regular expression.
    query_params: List of parameters (represented by parameter name as string)
        that will be used in the query string.
    path_params: Set of parameters (represented by parameter name as string)
        that will be used in the base URL path.
    param_types: Map from method parameter name (string) to parameter type. Type
        can be any valid JSON schema type; valid values are 'any', 'array',
        'boolean', 'integer', 'number', 'object', or 'string'. Reference:
        http://tools.ietf.org/html/draft-zyp-json-schema-03#section-5.1
    enum_params: Map from method parameter name (string) to list of strings,
       where each list of strings is the list of acceptable enum values.
  """

  def __init__(self, method_desc):
    """Constructor for ResourceMethodParameters.

    Sets default values and defers to set_parameters to populate.

    Args:
      method_desc: Dictionary with metadata describing an API method. Value
          comes from the dictionary of methods stored in the 'methods' key in
          the deserialized discovery document.
    """
    self.argmap = {}
    self.required_params = []
    self.repeated_params = []
    self.pattern_params = {}
    self.query_params = []
    # TODO(dhermes): Change path_params to a list if the extra URITEMPLATE
    #                parsing is gotten rid of.
    self.path_params = set()
    self.param_types = {}
    self.enum_params = {}

    self.set_parameters(method_desc)

  def set_parameters(self, method_desc):
    """Populates maps and lists based on method description.

    Iterates through each parameter for the method and parses the values from
    the parameter dictionary.

    Args:
      method_desc: Dictionary with metadata describing an API method. Value
          comes from the dictionary of methods stored in the 'methods' key in
          the deserialized discovery document.
    """
    for arg, desc in list(method_desc.get('parameters', {}).items()):
      param = key2param(arg)
      self.argmap[param] = arg

      if desc.get('pattern'):
        self.pattern_params[param] = desc['pattern']
      if desc.get('enum'):
        self.enum_params[param] = desc['enum']
      if desc.get('required'):
        self.required_params.append(param)
      if desc.get('repeated'):
        self.repeated_params.append(param)
      if desc.get('location') == 'query':
        self.query_params.append(param)
      if desc.get('location') == 'path':
        self.path_params.add(param)
      self.param_types[param] = desc.get('type', 'string')

    # TODO(dhermes): Determine if this is still necessary. Discovery based APIs
    #                should have all path parameters already marked with
    #                'location: path'.
    for match in URITEMPLATE.finditer(method_desc['path']):
      for namematch in VARNAME.finditer(match.group(0)):
        name = key2param(namematch.group(0))
        self.path_params.add(name)
        if name in self.query_params:
          self.query_params.remove(name)


def createMethod(methodName, methodDesc, rootDesc, schema):
  """Creates a method for attaching to a Resource.

  Args:
    methodName: string, name of the method to use.
    methodDesc: object, fragment of deserialized discovery document that
      describes the method.
    rootDesc: object, the entire deserialized discovery document.
    schema: object, mapping of schema names to schema descriptions.
  """
  methodName = fix_method_name(methodName)
  (pathUrl, httpMethod, methodId, accept,
   maxSize, mediaPathUrl) = _fix_up_method_description(methodDesc, rootDesc)

  parameters = ResourceMethodParameters(methodDesc)

  def method(self, **kwargs):
    # Don't bother with doc string, it will be over-written by createMethod.

    for name in list(kwargs.keys()):
      if name not in parameters.argmap:
        raise TypeError('Got an unexpected keyword argument "%s"' % name)

    # Remove args that have a value of None.
    keys = list(kwargs.keys())
    for name in keys:
      if kwargs[name] is None:
        del kwargs[name]

    for name in parameters.required_params:
      if name not in kwargs:
        raise TypeError('Missing required parameter "%s"' % name)

    for name, regex in list(parameters.pattern_params.items()):
      if name in kwargs:
        if isinstance(kwargs[name], str):
          pvalues = [kwargs[name]]
        else:
          pvalues = kwargs[name]
        for pvalue in pvalues:
          if re.match(regex, pvalue) is None:
            raise TypeError(
                'Parameter "%s" value "%s" does not match the pattern "%s"' %
                (name, pvalue, regex))

    for name, enums in list(parameters.enum_params.items()):
      if name in kwargs:
        # We need to handle the case of a repeated enum
        # name differently, since we want to handle both
        # arg='value' and arg=['value1', 'value2']
        if (name in parameters.repeated_params and
            not isinstance(kwargs[name], str)):
          values = kwargs[name]
        else:
          values = [kwargs[name]]
        for value in values:
          if value not in enums:
            raise TypeError(
                'Parameter "%s" value "%s" is not an allowed value in "%s"' %
                (name, value, str(enums)))

    actual_query_params = {}
    actual_path_params = {}
    for key, value in list(kwargs.items()):
      to_type = parameters.param_types.get(key, 'string')
      # For repeated parameters we cast each member of the list.
      if key in parameters.repeated_params and type(value) == type([]):
        cast_value = [_cast(x, to_type) for x in value]
      else:
        cast_value = _cast(value, to_type)
      if key in parameters.query_params:
        actual_query_params[parameters.argmap[key]] = cast_value
      if key in parameters.path_params:
        actual_path_params[parameters.argmap[key]] = cast_value
    body_value = kwargs.get('body', None)
    media_filename = kwargs.get('media_body', None)

    if self._developerKey:
      actual_query_params['key'] = self._developerKey

    model = self._model
    if methodName.endswith('_media'):
      model = MediaModel()
    elif 'response' not in methodDesc:
      model = RawModel()

    headers = {}
    headers, params, query, body = model.request(headers,
        actual_path_params, actual_query_params, body_value)

    expanded_url = uritemplate.expand(pathUrl, params)
    url = urllib.parse.urljoin(self._baseUrl, expanded_url + query)

    resumable = None
    multipart_boundary = ''

    if media_filename:
      # Ensure we end up with a valid MediaUpload object.
      if isinstance(media_filename, str):
        (media_mime_type, encoding) = mimetypes.guess_type(media_filename)
        if media_mime_type is None:
          raise UnknownFileType(media_filename)
        if not mimeparse.best_match([media_mime_type], ','.join(accept)):
          raise UnacceptableMimeTypeError(media_mime_type)
        media_upload = MediaFileUpload(media_filename,
                                       mimetype=media_mime_type)
      elif isinstance(media_filename, MediaUpload):
        media_upload = media_filename
      else:
        raise TypeError('media_filename must be str or MediaUpload.')

      # Check the maxSize
      if maxSize > 0 and media_upload.size() > maxSize:
        raise MediaUploadSizeError("Media larger than: %s" % maxSize)

      # Use the media path uri for media uploads
      expanded_url = uritemplate.expand(mediaPathUrl, params)
      url = urllib.parse.urljoin(self._baseUrl, expanded_url + query)
      if media_upload.resumable():
        url = _add_query_parameter(url, 'uploadType', 'resumable')

      if media_upload.resumable():
        # This is all we need to do for resumable, if the body exists it gets
        # sent in the first request, otherwise an empty body is sent.
        resumable = media_upload
      else:
        # A non-resumable upload
        if body is None:
          # This is a simple media upload
          headers['content-type'] = media_upload.mimetype()
          body = media_upload.getbytes(0, media_upload.size())
          url = _add_query_parameter(url, 'uploadType', 'media')
        else:
          # This is a multipart/related upload.
          msgRoot = MIMEMultipart('related')
          # msgRoot should not write out it's own headers
          setattr(msgRoot, '_write_headers', lambda self: None)

          # attach the body as one part
          msg = MIMENonMultipart(*headers['content-type'].split('/'))
          msg.set_payload(body)
          msgRoot.attach(msg)

          # attach the media as the second part
          msg = MIMENonMultipart(*media_upload.mimetype().split('/'))
          msg['Content-Transfer-Encoding'] = 'binary'

          payload = media_upload.getbytes(0, media_upload.size())
          msg.set_payload(payload)
          msgRoot.attach(msg)
          body = msgRoot.as_string()

          multipart_boundary = msgRoot.get_boundary()
          headers['content-type'] = ('multipart/related; '
                                     'boundary="%s"') % multipart_boundary
          url = _add_query_parameter(url, 'uploadType', 'multipart')

    logger.info('URL being requested: %s' % url)
    return self._requestBuilder(self._http,
                                model.response,
                                url,
                                method=httpMethod,
                                body=body,
                                headers=headers,
                                methodId=methodId,
                                resumable=resumable)

  docs = [methodDesc.get('description', DEFAULT_METHOD_DOC), '\n\n']
  if len(parameters.argmap) > 0:
    docs.append('Args:\n')

  # Skip undocumented params and params common to all methods.
  skip_parameters = list(rootDesc.get('parameters', {}).keys())
  skip_parameters.extend(STACK_QUERY_PARAMETERS)

  all_args = list(parameters.argmap.keys())
  args_ordered = [key2param(s) for s in methodDesc.get('parameterOrder', [])]

  # Move body to the front of the line.
  if 'body' in all_args:
    args_ordered.append('body')

  for name in all_args:
    if name not in args_ordered:
      args_ordered.append(name)

  for arg in args_ordered:
    if arg in skip_parameters:
      continue

    repeated = ''
    if arg in parameters.repeated_params:
      repeated = ' (repeated)'
    required = ''
    if arg in parameters.required_params:
      required = ' (required)'
    paramdesc = methodDesc['parameters'][parameters.argmap[arg]]
    paramdoc = paramdesc.get('description', 'A parameter')
    if '$ref' in paramdesc:
      docs.append(
          ('  %s: object, %s%s%s\n    The object takes the'
          ' form of:\n\n%s\n\n') % (arg, paramdoc, required, repeated,
            schema.prettyPrintByName(paramdesc['$ref'])))
    else:
      paramtype = paramdesc.get('type', 'string')
      docs.append('  %s: %s, %s%s%s\n' % (arg, paramtype, paramdoc, required,
                                          repeated))
    enum = paramdesc.get('enum', [])
    enumDesc = paramdesc.get('enumDescriptions', [])
    if enum and enumDesc:
      docs.append('    Allowed values\n')
      for (name, desc) in zip(enum, enumDesc):
        docs.append('      %s - %s\n' % (name, desc))
  if 'response' in methodDesc:
    if methodName.endswith('_media'):
      docs.append('\nReturns:\n  The media object as a string.\n\n    ')
    else:
      docs.append('\nReturns:\n  An object of the form:\n\n    ')
      docs.append(schema.prettyPrintSchema(methodDesc['response']))

  setattr(method, '__doc__', ''.join(docs))
  return (methodName, method)


def createNextMethod(methodName):
  """Creates any _next methods for attaching to a Resource.

  The _next methods allow for easy iteration through list() responses.

  Args:
    methodName: string, name of the method to use.
  """
  methodName = fix_method_name(methodName)

  def methodNext(self, previous_request, previous_response):
    """Retrieves the next page of results.

Args:
  previous_request: The request for the previous page. (required)
  previous_response: The response from the request for the previous page. (required)

Returns:
  A request object that you can call 'execute()' on to request the next
  page. Returns None if there are no more items in the collection.
    """
    # Retrieve nextPageToken from previous_response
    # Use as pageToken in previous_request to create new request.

    if 'nextPageToken' not in previous_response:
      return None

    request = copy.copy(previous_request)

    pageToken = previous_response['nextPageToken']
    parsed = list(urllib.parse.urlparse(request.uri))
    q = parse_qsl(parsed[4])

    # Find and remove old 'pageToken' value from URI
    newq = [(key, value) for (key, value) in q if key != 'pageToken']
    newq.append(('pageToken', pageToken))
    parsed[4] = urllib.parse.urlencode(newq)
    uri = urllib.parse.urlunparse(parsed)

    request.uri = uri

    logger.info('URL being requested: %s' % uri)

    return request

  return (methodName, methodNext)


class Resource(object):
  """A class for interacting with a resource."""

  def __init__(self, http, baseUrl, model, requestBuilder, developerKey,
               resourceDesc, rootDesc, schema):
    """Build a Resource from the API description.

    Args:
      http: httplib2.Http, Object to make http requests with.
      baseUrl: string, base URL for the API. All requests are relative to this
          URI.
      model: apiclient.Model, converts to and from the wire format.
      requestBuilder: class or callable that instantiates an
          apiclient.HttpRequest object.
      developerKey: string, key obtained from
          https://code.google.com/apis/console
      resourceDesc: object, section of deserialized discovery document that
          describes a resource. Note that the top level discovery document
          is considered a resource.
      rootDesc: object, the entire deserialized discovery document.
      schema: object, mapping of schema names to schema descriptions.
    """
    self._dynamic_attrs = []

    self._http = http
    self._baseUrl = baseUrl
    self._model = model
    self._developerKey = developerKey
    self._requestBuilder = requestBuilder
    self._resourceDesc = resourceDesc
    self._rootDesc = rootDesc
    self._schema = schema

    self._set_service_methods()

  def _set_dynamic_attr(self, attr_name, value):
    """Sets an instance attribute and tracks it in a list of dynamic attributes.

    Args:
      attr_name: string; The name of the attribute to be set
      value: The value being set on the object and tracked in the dynamic cache.
    """
    self._dynamic_attrs.append(attr_name)
    self.__dict__[attr_name] = value

  def __getstate__(self):
    """Trim the state down to something that can be pickled.

    Uses the fact that the instance variable _dynamic_attrs holds attrs that
    will be wiped and restored on pickle serialization.
    """
    state_dict = copy.copy(self.__dict__)
    for dynamic_attr in self._dynamic_attrs:
      del state_dict[dynamic_attr]
    del state_dict['_dynamic_attrs']
    return state_dict

  def __setstate__(self, state):
    """Reconstitute the state of the object from being pickled.

    Uses the fact that the instance variable _dynamic_attrs holds attrs that
    will be wiped and restored on pickle serialization.
    """
    self.__dict__.update(state)
    self._dynamic_attrs = []
    self._set_service_methods()

  def _set_service_methods(self):
    self._add_basic_methods(self._resourceDesc, self._rootDesc, self._schema)
    self._add_nested_resources(self._resourceDesc, self._rootDesc, self._schema)
    self._add_next_methods(self._resourceDesc, self._schema)

  def _add_basic_methods(self, resourceDesc, rootDesc, schema):
    # Add basic methods to Resource
    if 'methods' in resourceDesc:
      for methodName, methodDesc in list(resourceDesc['methods'].items()):
        fixedMethodName, method = createMethod(
            methodName, methodDesc, rootDesc, schema)
        self._set_dynamic_attr(fixedMethodName,
                               method.__get__(self, self.__class__))
        # Add in _media methods. The functionality of the attached method will
        # change when it sees that the method name ends in _media.
        if methodDesc.get('supportsMediaDownload', False):
          fixedMethodName, method = createMethod(
              methodName + '_media', methodDesc, rootDesc, schema)
          self._set_dynamic_attr(fixedMethodName,
                                 method.__get__(self, self.__class__))

  def _add_nested_resources(self, resourceDesc, rootDesc, schema):
    # Add in nested resources
    if 'resources' in resourceDesc:

      def createResourceMethod(methodName, methodDesc):
        """Create a method on the Resource to access a nested Resource.

        Args:
          methodName: string, name of the method to use.
          methodDesc: object, fragment of deserialized discovery document that
            describes the method.
        """
        methodName = fix_method_name(methodName)

        def methodResource(self):
          return Resource(http=self._http, baseUrl=self._baseUrl,
                          model=self._model, developerKey=self._developerKey,
                          requestBuilder=self._requestBuilder,
                          resourceDesc=methodDesc, rootDesc=rootDesc,
                          schema=schema)

        setattr(methodResource, '__doc__', 'A collection resource.')
        setattr(methodResource, '__is_resource__', True)

        return (methodName, methodResource)

      for methodName, methodDesc in list(resourceDesc['resources'].items()):
        fixedMethodName, method = createResourceMethod(methodName, methodDesc)
        self._set_dynamic_attr(fixedMethodName,
                               method.__get__(self, self.__class__))

  def _add_next_methods(self, resourceDesc, schema):
    # Add _next() methods
    # Look for response bodies in schema that contain nextPageToken, and methods
    # that take a pageToken parameter.
    if 'methods' in resourceDesc:
      for methodName, methodDesc in list(resourceDesc['methods'].items()):
        if 'response' in methodDesc:
          responseSchema = methodDesc['response']
          if '$ref' in responseSchema:
            responseSchema = schema.get(responseSchema['$ref'])
          hasNextPageToken = 'nextPageToken' in responseSchema.get('properties',
                                                                   {})
          hasPageToken = 'pageToken' in methodDesc.get('parameters', {})
          if hasNextPageToken and hasPageToken:
            fixedMethodName, method = createNextMethod(methodName + '_next')
            self._set_dynamic_attr(fixedMethodName,
                                   method.__get__(self, self.__class__))



from ..exceptions import *

import os
import tempfile
import subprocess

websnap_cmd = [ '/opt/phantomjs/bin/phantomjs', '/opt/phantomjs/webtopng.js' ]

def run_cmd(args):
    fp = tempfile.TemporaryFile()
    nfd = os.open("/dev/null", os.O_RDONLY)
    retval = subprocess.call(args, close_fds=True, stderr=fp.fileno(), stdout=fp.fileno(), stdin=nfd)
    os.close(nfd)
    fp.flush()
    fp.seek(0)
    bufl = fp.readlines()
    fp.close()
    buf = ""
    for l in bufl:
        buf += l
    if retval:
        raise CmdError(retval, buf, cmd=args)
    return retval

def run(outfile, web_url):
    cmd = websnap_cmd + [web_url, outfile]
    try:
        run_cmd(cmd)
    except Exception as e:
        try:
            os.unlink(outfile)
        except:
            pass
        raise e
    return outfile



from django.http import HttpResponse
from django.shortcuts import render
from django.db.models.query import QuerySet

from rest import settings
from version import VERSION
import time
from auditlog.models import PersistentLog
from . import model as ms
from . import collection as cs
from . import json as js
from . import csv
from . import excel
# from . import profiler


def restStatus(request, status, data={}, **kwargs):
    if isinstance(data, str):
        if status:
            data = dict(message=data)
        else:
            data = dict(error=data)
    data = data.copy()
    data.update(kwargs)
    data["status"] = status
    return restResult(request, data)


def restSuccessNoContent(request):
    return HttpResponse(status=204)


# @profiler.timeit
def restGet(request, obj, fields=None, extra=[], exclude=[], recurse_into=[], **kwargs):
    data = ms.to_dict(obj, fields, extra, exclude, recurse_into)
    if not kwargs.get("return_httpresponse", True):
        return data
    data = dict(data=data, status=kwargs.get("status", True))
    data["datetime"] = int(time.time())
    data["elapsed"] = get_request_elapsed(request)
    return restResult(request, data)

# @profiler.timeit
def restList(request, qset, size=25, start=0, sort=None, fields=None, extra=[], exclude=[], recurse_into=[], **kwargs):
    count = 0
    if isinstance(qset, QuerySet):
        count = qset.count()
    elif isinstance(qset, list):
        count = len(qset)
    start = request.DATA.get("start", start, field_type=int)
    size = request.DATA.get("size", size, field_type=int)
    sort = request.DATA.get("sort", sort)
    data = cs.to_list(qset, sort, size, start, fields, extra, exclude, recurse_into)
    data["status"] = kwargs.get("status", True)
    data["datetime"] = int(time.time())
    data["elapsed"] = get_request_elapsed(request)
    data["count"] = count
    if "response_params" in kwargs:
        for k, v in kwargs.get("response_params").items():
            data[k] = v
    return restResult(request, data)


def restResult(request, data, status=200):
    if "status" not in data:
        data["status"] = True
        data["datetime"] = int(time.time())
        data["elapsed"] = get_request_elapsed(request)
    accept_list = parse_accept_list(request)
    restResponseLog(request, data)
    if 'application/json' in accept_list:
        return HttpResponse(js.serialize(data), content_type="application/json", status=status)
    elif 'text/html' in accept_list:
        return render_webview(request, data, status)
    return HttpResponse(js.prettyJSON(data), content_type="text/plain", status=status)


restReturn = restResult  # legacy support


def restResponseLog(request, resp):
    # render response
    if hasattr(request, "rest_debug") and request.rest_debug:
        if settings.DEBUG_REST_NO_LISTS and "count" in resp:
            if "count" in resp:
                if settings.DEBUG_REST_NO_LIST == "example":
                    sanatized = dict(count=resp.get("count"), size=resp.get("size"), data=resp.get("data", [])[:1])
                else:
                    sanatized = dict(count=resp.get("count"), size=resp.get("size"))
                PersistentLog.log(sanatized, 0, request, "rest", action="response")
            else:
                PersistentLog.log(resp, 0, request, "rest", action="response")
        else:
            PersistentLog.log(resp, 0, request, "rest", action="response")


def restPermissionDenied(request, error="permission denied", error_code=403, status=403,
                         component=None, component_id=None):
    # from rest import helpers as rh
    # rh.log_error(error)
    return restResult(request, dict(status=False, error=error, error_code=error_code), status=status)


def restNotFound(request):
    return restStatus(request, False, error="not found", error_code=404)


def restOK(request):
    return restStatus(request, True)


def restError(request, error, error_code=None):
    return restStatus(request, False, error=error, error_code=error_code)


def restJSON(request, qset, fields, name, size=10000, localize=None):
    if fields:
        data = cs.to_list(qset, fields=fields, size=size)
    else:
        data = cs.serialize(qset, request.DATA.get("graph"), size=size)
    response = HttpResponse(js.prettyJSON(data), content_type="application/json")
    response['Content-Disposition'] = 'attachment; filename="{0}"'.format(name)
    return response


def restFlat(request, qset, fields, name, size=10000, header_cols=None, localize=None):
    return HttpResponse(js.prettyJSON(list(qset.values_list(*fields))), content_type="application/json", status=200)


def restCSV(request, qset, fields, name, size=10000, header_cols=None, localize=None):
    if size > 1200 and qset.count() > 1200:
        # large data set lets stream
        qset = qset[:size]
        response = csv.generateCSVStream(qset, fields, name, localize=localize)
        response['Cache-Control'] = 'no-cache'
        response['Content-Disposition'] = f'attachment; filename="{name}"'
    else:
        response = HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = f'attachment; filename="{name}"'
        qset = qset[:size]
        csv.generateCSV(qset, fields, name, output=response, header_cols=header_cols, localize=localize)
    return response


def restExcel(request, qset, fields, name, size=10000, localize=None, **kwargs):
    return excel.qsetToExcel(request, qset[:size], fields, name)


def parse_accept_list(request):
    if request and hasattr(request, "DATA") and request.DATA.get('_type', None) is not None:
        accept_list = [request.DATA.get('_type')]
    elif request and 'HTTP_ACCEPT' in request.META:
        accept_list = request.META["HTTP_ACCEPT"].split(',')
    elif request and 'HTTP_ACCEPT_ENCODING' in request.META:
        accept_list = request.META["HTTP_ACCEPT_ENCODING"].split(',')
    else:
        accept_list = []
    return accept_list


def get_request_elapsed(request):
    if request is not None:
        return int((time.perf_counter() - request._started)*1000)
    return 0


def render_webview(request, data, status=200):
    output = js.prettyJSON(data)
    req_dict = request.DATA.asDict()
    req_out = js.prettyJSON(req_dict)
    path = request.path
    parts = path.split('/')
    last_pk = parts.pop()
    if last_pk and last_pk.isdigit():
        path = "/".join(parts)
    else:
        last_pk = ""

    is_debug = settings.get("DEBUG", False) or settings.get("REST_DEBUGGER", False)
    rest_help = {}
    if is_debug and hasattr(request, "rest_class"):
        rest_help = request.rest_class.get_rest_help()
        rest_help = js.prettyJSON(rest_help)
    context = {
        "help": rest_help,
        "input": req_dict,
        "pk": last_pk,
        "output_dict": data,
        "output": output,
        "req_out": req_out,
        "path": path,
        "method": request.method,
        "debug": is_debug,
        "request": request,
        "version": VERSION,
        "disclaimer": settings.REST_DISCLAIMER
    }
    return render(request, "rest_html.html", context, status=status)



import json
from base64 import urlsafe_b64encode

try:
    from fido2 import webauthn
    import fido2
    from fido2.server import Fido2Server
    from fido2.utils import websafe_decode, websafe_encode
    from fido2 import cbor
    fido2.features.webauthn_json_mapping.enabled = True
except Exception:
    pass

from rest import settings
from rest import helpers as rh
from account.models import UserPassKey


from objict import objict
# platform == fires apple keychain
# cross-platform = tries for bluetooh
FIDO_KEY_ATTACHMENT = settings.get("FIDO_KEY_ATTACHMENT", "cross-platform")

FIDO_SERVER_ID = settings.get("FIDO_SERVER_ID", settings.SERVER_NAME)
FIDO_SERVER_NAME = settings.get("FIDO_SERVER_NAME", settings.SITE_LABEL)


def verify_origin(id):
    return True


def getServer(request=None, rp_id=FIDO_SERVER_ID, rp_name=FIDO_SERVER_NAME):
    if request is not None:
        rp_id = request.DATA.get(["rp_id", "fido_server_id"], rp_id)
        rp_name = request.DATA.get(["rp_name", "fido_server_name"], rp_name)
    rp = webauthn.PublicKeyCredentialRpEntity(id=rp_id, name=rp_name)
    return Fido2Server(rp, verify_origin=verify_origin)


def registerBegin(member, request, attachment=FIDO_KEY_ATTACHMENT):
    """
    data = CredentialCreationOptions
    state = {'challenge': '4yZmyZmnWP11t7g1S151oVgL0Vw0AU9GegTYJM2_928', 'user_verification': None}
    """

    server = getServer(request)
    reg_data = dict(
        id=rh.toBase64(member.getUUID()),
        name=member.username,
        displayName=member.display_name)
    data, state = server.register_begin(
        reg_data,
        authenticator_attachment=attachment,
        resident_key_requirement=webauthn.ResidentKeyRequirement.PREFERRED)
    rp = objict(server.rp)
    data = objict.fromdict(dict(data))
    data.excludeCredentials = getUserCredentials(member, websafe=True)
    rh.debug("data", data)
    # rh.debug("registerBegin", rp, server.rp.id_hash)
    return data, state, rp


def registerComplete(request, fido2_state, rp_id):
    credentials = request.DATA.get("credentials")
    # rh.debug("registerComplete", fido2_state, rp_id)
    server = getServer(request, rp_id)
    rp = objict(server.rp)
    # rh.debug("registerBegin", rp, server.rp.id_hash)

    auth_data = server.register_complete(
        fido2_state,
        response=credentials
    )

    user_key = UserPassKey(
        uuid=credentials.id,
        name=request.DATA.get("key_name", ""),
        rp_id=rp_id,
        member=request.member,
        platform=request.DATA.getUserAgentPlatform(),
        token=websafe_encode(auth_data.credential_data))
    # Store `auth_data.credential_data` in your database associated with the user
    user_key.save()
    return user_key


def getUserCredentials(member, websafe=False):
    if member is None:
        return []
    creds = [webauthn.AttestedCredentialData(websafe_decode(uk.token)) for uk in member.passkeys.all()]
    if websafe:
        return [dict(
            type="public-key",
            id=rh.toBase64(acd.credential_id)) for acd in creds]
    return creds


def authBegin(request):
    server = getServer(request)
    stored_credentials = getUserCredentials(request.member)
    auth_data, state = server.authenticate_begin(stored_credentials)
    return auth_data, state, objict(server.rp)


def authComplete(request, fido2_state, rp_id):
    credential = request.DATA.get("credential")
    upk = UserPassKey.objects.filter(uuid=credential.id, is_enabled=1).last()
    if upk is None:
        raise Exception("PassKey not found on host.")
    stored_credentials = [webauthn.AttestedCredentialData(websafe_decode(upk.token))]
    server = getServer(request, rp_id)
    cred = server.authenticate_complete(
        fido2_state,
        credentials=stored_credentials,
        response=credential)
    request.member = upk.member
    request.member.canLogin(request)  # throws exception if cannot login
    request.member.login(request=request)
    request.member.log(
        "passkey_login", "passkey login succesful", 
        request, method="login", level=7)
    upk.touch()
    return upk



import hashlib

def toHex(value):
    return binascii.hexlify(value).upper().decode('utf-8')

def hexToByteArray(value):
    return bytearray.fromhex(value)

def toString(value):
    if isinstance(value, bytes):
        value = value.decode()
    elif isinstance(value, bytearray):
        value = value.decode("utf-8")
    return value

def toBytes(value):
    if isinstance(value, str):
        value = value.encode("utf-8")
    elif isinstance(value, bytearray):
        value = bytes(value)
    return value


def _xor_idstr(data, encrypt):
    lastc = 51
    cnt = 126
    r = []
    data = toString(data)
    for i in range(len(data)):
        r.append(chr(ord(data[i]) ^ lastc ^ ((cnt*7)%256) ^ (11 * (2**(lastc % 5)))  ))
        if encrypt:
            lastc = lastc ^ ord(data[i])
        else:
            lastc = lastc ^ ord(r[i])
    return r

def _rc4_round(data, key, encrypt, discard):
    # setup sbox
    x = 0
    box = list(range(256))
    key = toString(key)
    for i in range(256):
        x = (x + box[i] + ord(key[i % len(key)])) % 256
        box[i], box[x] = box[x], box[i]
    x = 0
    y = 0
    out = []

    # discard first rounds
    for i in range(discard):
        x = (x + 1) % 256
        y = (y + box[x]) % 256
        box[x], box[y] = box[y], box[x]

    # pre process
    if encrypt:
        data = _xor_idstr(data, encrypt)
    else:
        data.reverse()

    # rc4 crypt
    for char in data:
        x = (x + 1) % 256
        y = (y + box[x]) % 256
        box[x], box[y] = box[y], box[x]
        out.append(chr(ord(char) ^ box[(box[x] + box[y]) % 256]))

    # post process
    if encrypt:
        out.reverse()
    else:
        out = _xor_idstr(out, encrypt)

    return out

def crypt(data, key, encrypt):
    key = hashlib.md5(key).digest()

    data = list(data)
    out = _rc4_round(data, key, encrypt, encrypt and 871 or 917)
    out = _rc4_round(out, key, encrypt, encrypt and 917 or 871)

    return ''.join(out)



from rest import helpers as rh
from rest.log import getLogger
import time


def run_example(task):
    # a background task has been pushed to this call
    logger = getLogger("wiki", "wiki.log")
    task.log("starting task")
    time.sleep(10)
    task.log("finished my task")
    logger.info("task completed")





from django.db import models as dm
from rest import models as rm


class PageRevision(dm.Model, rm.RestModel):
    created = dm.DateTimeField(auto_now_add=True, editable=False)

    owner = dm.ForeignKey(
        "account.Member", related_name="+",
        blank=True, null=True, default=None,
        on_delete=dm.SET_NULL)

    group = dm.ForeignKey(
        "account.Group", related_name="+", 
        blank=True, null=True, default=None,
        on_delete=dm.SET_NULL)

    entry = dm.ForeignKey(
        "wiki.Page", related_name="revisions", 
        default=None, null=True, blank=True,
        on_delete=dm.CASCADE)

    title = dm.CharField(max_length=255)
    slug = dm.SlugField()

    body = dm.TextField(blank=True)



from rest import settings
from .models import PersistentLog

DEBUG_REST_ALL = settings.get("DEBUG_REST_ALL", False)
DEBUG_REST_END_POINTS = settings.get("DEBUG_REST_END_POINTS", [])
IGNORE_REST_END_POINTS = settings.get("IGNORE_REST_END_POINTS", [])
LOG_REST_PREFIX = settings.get("REST_PREFIX", "api/")
if not LOG_REST_PREFIX.startswith("/"):
    LOG_REST_PREFIX = f"/{LOG_REST_PREFIX}"


def checkRestDebug(request):
    if DEBUG_REST_ALL:
        return True
    for ep in DEBUG_REST_END_POINTS:
        if request.path.startswith(ep):
            return not checkRestIgnore(request)
    return False


def checkRestIgnore(request):
    for ep in IGNORE_REST_END_POINTS:
        if isinstance(ep, tuple):
            method, ep = ep
            return request.method == method and request.path.startswith(ep)
        if request.path.startswith(ep):
            return True
    return False


class LogRequest(object):
    last_request = None

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        return self.process_request(request)

    def process_request(self, request):
        # LogRequest.last_request = request
        request.rest_debug = False
        if request.path.startswith(LOG_REST_PREFIX) or request.path.startswith("/rpc/"):
            request.rest_debug = checkRestDebug(request)
            if request.rest_debug:
                request.DATA.log()
        response = self.get_response(request)
        return response




from django.contrib import admin
from .models import *
from django import forms

from django.urls import include, re_path, reverse

class PersistentLogAdmin(admin.ModelAdmin):
    search_fields = ["group__name", "user__username", "component", "session__session_id", "session__ip", "message", "request_path"]
    list_display = ('when', 'tid', 'request_method', 'request_path', 'component', 'action')

    def get_readonly_fields(self, request, obj = None):
        return ('message', 'level', 'user', 'group', 'session', 'request_path', 'request_method', 'component', 'action', 'pkey')

    def has_delete_permission(self, request, obj=None):
        return False

    def session_link(self, obj):
        if obj.session is None:
            return "n/a"
        url = reverse('admin:sessionlog_sessionlog_change', args=(obj.session.pk,))
        return '<a href="{0}">{1}</a>'.format(url, obj.session.session_id)
    session_link.allow_tags = True

admin.site.register(PersistentLog, PersistentLogAdmin)




from . import views

import version
import time
from datetime import datetime
import platform
import subprocess
from . import joke
from . import helpers

import django
from django.conf import settings


from .decorators import url

try:
    import psutil
except Exception:
    print("no psutil")


URL_PREFIX = ""

SOFTWARE_VERSIONS = getattr(settings, 'SOFTWARE_VERSIONS', None)
ALLOW_PUBLIC_SYS_INFO = getattr(settings, 'ALLOW_PUBLIC_SYS_INFO', False)
# SOFTWARE_VERSIONS_ACTUAL = {}


@url(r'^version$')
def on_get_version(request):
    return views.restStatus(request, True, {"data": version.VERSION, "ip": request.ip})


@url(r'^versions$')
def on_get_version(request):
    from rest import __version__ as restit_version
    versions = dict(
        project=version.VERSION,
        restit=restit_version)
    if request.DATA.get("detailed") and SOFTWARE_VERSIONS:
        getVersions(versions)
    return views.restStatus(request, True, {"data": versions})


@url('test/session')
def on_get_my_session(request):
    echo = request.DATA.get("echo", None)
    if echo:
        request.session["echo"] = echo
    return views.restReturn(request, dict(echo=request.session.get("echo", "not set")))


@url(r'^joke$')
def on_get_joke(request):
    return views.restGet(request, {"joke": joke.getRandomJoke()})


def getTcpEstablishedCount():
    return len(getTcpEstablished())


def getTcpEstablished():
    cons = psutil.net_connections(kind="tcp")
    established = []
    for c in cons:
        if c.status == "ESTABLISHED":
            established.append(c)
    return established


@url(r'^system/info$')
def on_get_system_info(request):
    if not ALLOW_PUBLIC_SYS_INFO and not request.user.is_authenticated:
        return views.restPermissionDenied(request, "system info is not public", 565)
    mem = psutil.virtual_memory()
    disk = psutil.disk_usage('/')
    net = psutil.net_io_counters()

    out = {
        "time": time.time(),
        "datetime": str(datetime.now()),
        "version": version.VERSION,
        "os": {
            "system": platform.system(),
            "version": platform.version(),
            "hostname": platform.node(),
            "release": platform.release(),
            "processor": platform.processor(),
            "machine": platform.machine()
        },
        "boot_time": psutil.boot_time(),
        "cpu_load": psutil.cpu_percent(),
        "cpus_load": psutil.cpu_percent(percpu=True),
        "memory": {
            "total": mem.total,
            "used": mem.used,
            "available": mem.available,
            "percent": mem.percent
        },
        "disk": {
            "total": disk.total,
            "used": disk.used,
            "free": disk.free,
            "percent": disk.percent
        }, 
        "users": psutil.users()
    }

    try:
        out["cpu"] = {
            "count": psutil.cpu_count(),
            "freq": psutil.cpu_freq(),
        }
    except Exception:
        pass

    try:
        out["network"] = {
            "tcp_cons": getTcpEstablishedCount(),
            "bytes_sent": net.bytes_sent,
            "bytes_recv": net.bytes_recv,
            "packets_sent": net.packets_sent,
            "packets_recv": net.packets_recv,
            "errin": net.errin,
            "errout": net.errout,
            "dropin": net.dropin,
            "dropout": net.dropout
        }
    except Exception:
        pass

    if request.DATA.get("versions") and SOFTWARE_VERSIONS:
        out["versions"] = getVersions()
    if request.DATA.get("blocked"):
        out["blocked"] = helpers.getBlockedHosts()
    return views.restGet(request, out)


def safe_cmd(cmd, *args):
    try:
        cmd_args = [cmd]
        if len(args):
            cmd_args.extend(list(args))
        return helpers.toString(subprocess.check_output(cmd_args, shell=True).strip())
    except Exception as err:
        return str(err)
        # print( str(err))
    return None


def getVersions(out=None):
    if out is None:
        out = {}
    for key in SOFTWARE_VERSIONS:
        if key == "django":
            out[key] = django.__version__
        else:
            out[key] = safe_cmd(SOFTWARE_VERSIONS[key])
    return out



from django.db import models
# from django.db.models import Q
from rest import settings
import time
from objict import objict

from rest.models import RestModel, MetaDataModel, MetaDataBase, RestValidationError


DEFAULT_ROLE = settings.get("MEMBERSHIP_DEFAULT_ROLE", "guest")


class Membership(models.Model, RestModel, MetaDataModel):
    class RestMeta:
        CAN_DELETE = True
        SEARCH_FIELDS = ["role", "member__username", "member__first_name", "member__last_name", "member__email"]
        LIST_DEFAULT_FILTERS = {
            "state__gte": 0
        }
        LIST_PARENT_KINDS = []
        VIEW_PERMS = ["view_all_groups", "manage_members", "manage_group", "manage_users", "manage_groups", "owner"]
        VIEW_PERMS_MEMBER_FIELD = "member"
        SAVE_PERMS = ["manage_groups", "create_groups", "manage_users", "manage_members"]
        CREATE_PERMS = ["manage_groups", "create_groups", "manage_users", "manage_members"]
        SEARCH_TERMS = [
            ("username", "member__username"),
            ("email", "member__email"),
            ("first_name", "member__first_name"),
            ("last_name", "member__last_name"),
            ("last_activity", "member__last_activity#datetime"),
            ("created", "member__datejoined#datetime"),
            ("perms", "properties|permissions."),
            "role"]
        METADATA_FIELD_PROPERTIES = settings.MEMBERSHIP_METADATA_PROPERTIES
        GRAPHS = {
            "default": {
                "extra": ["metadata"],
                "fields": [
                    'id',
                    'created',
                    'role',
                    'status',
                    'state'
                ],
                "graphs": {
                    "member": "basic"
                },
            },
            "list": {
                "extra": ["metadata"],
                "fields": [
                    'id',
                    'created',
                    'role',
                    'status',
                    'state'
                ],
                "graphs": {
                    "member": "basic",
                    "group": "basic"
                },
            },
            "detailed": {
                "graphs": {
                    "self": "default",
                    "member": "detailed",
                }
            }
        }

    member = models.ForeignKey("account.Member", related_name="memberships", on_delete=models.CASCADE)
    group = models.ForeignKey("account.Group", related_name="memberships", on_delete=models.CASCADE)
    created = models.DateTimeField(auto_now_add=True)
    role = models.CharField(max_length=64, blank=True, null=True, default=DEFAULT_ROLE)
    state = models.IntegerField(default=0)

    @property
    def is_enabled(self):
        return self.state >= -10

    def set_action(self, value):
        if value == "resend_invite":
            self.sendInvite(self.getActiveRequest())

    def sendInvite(self, request=None, url=None, subject=None, site_logo=None, company_name=None):
        if request:
            powered_by = request.DATA.get("powered_by", True)
            subject = request.DATA.get("invite_subject")
            url = request.DATA.get("invite_url")
            site_logo = request.DATA.get("site_logo", settings.SITE_LOGO)
            company_name = request.DATA.get("company_name", settings.COMPANY_NAME)
        else:
            powered_by = True
            site_logo = settings.SITE_LOGO
            company_name = settings.COMPANY_NAME
        if url is None:
            raise RestValidationError("requires url", -1)
        if "?" not in url:
            url += "?jjj=1"
        if subject is None:
            subject = F"Invitation to {self.group.name}"
        is_new = not self.member.hasLoggedIn()
        btn_label = "ACCEPT"
        if is_new and url is not None:
            btn_label = "ACCEPT"
            expires = time.time() + 172800
            self.member.generateAuthCode(expires=expires)
            auth_token = objict(username=self.member.username, auth_token=self.member.auth_code)
            url = "{}&auth_code={}".format(url, auth_token.toBase64())
        self.member.sendInvite(
            subject, self.group, url=url, 
            msg=request.DATA.get("invite_msg", None), 
            POWERED_BY=powered_by,
            SITE_LOGO=site_logo,
            COMPANY_NAME=company_name,
            btn_label=btn_label)

    def set_state(self, value):
        if self.state != value:
            if value < -10:
                request = self.getActiveRequest()
                by_user = "system"
                if request:
                    by_user = request.member.username
                    request.member.auditLog(F"disabled {self.member.username} access to {self.group.name}", "membership_disabled")
                # we are disabling this member
                msg = F"{self.member.username} access to {self.group.name} disabled by {by_user}"
                self.auditLog(msg, "membership_disabled")
                self.member.auditLog(msg, "membership_disabled")
                self.state = value
            elif self.state < -10 and value >= -10:
                request = self.getActiveRequest()
                by_user = "system"
                if request:
                    by_user = request.member.username
                    request.member.auditLog(F"enabled {self.member.username} access to {self.group.name}", "membership_enabled")
                msg = F"{self.member.username} access to {self.group.name} enabled by {by_user}"
                self.auditLog(msg, "membership_enabled")
                self.member.auditLog(msg, "membership_enabled")
                self.state = value
            else:
                self.state = value
    
    def set_permissions(self, value):
        if isinstance(value, dict):
            self.setProperties(value, category="permissions")
        elif isinstance(value, list):
            for k in value:
                self.addPermission(k)

    def addPermission(self, perm):
        self.setProperty(perm, 1, "permissions")

    def removePermission(self, perm):
        self.setProperty(perm, None, "permissions")

    def clearPermissions(self):
        return self.properties.filter(category="permissions").delete()

    def getPermissions(self):
        return list(self.properties.filter(category="permissions", int_value__gt=0).values_list("key", flat=True))

    def hasPermission(self, perm):
        return self.hasPerm(perm)

    def hasPerm(self, perm):
        if not self.is_enabled:
            return False
        if isinstance(perm, list):
            for i in perm:
                if self.hasPerm(i):
                    return True
            return False
        return self.getProperty(perm, 0, "permissions", bool)

    def hasRole(self, role):
        if not self.is_enabled:
            return False
        if type(role) is list:
            return self.role in role
        return self.role == role

    # SUPPORT FOR LEGACY PERMS
    def set_perms(self, value):
        if isinstance(value, dict):
            for k, v in list(value.items()):
                if v in [1, "1", True, "true"]:
                    self.addPermission(k)
                elif v in [0, "0", False, "false"]:
                    self.removePermission(k)
        elif isinstance(value, list):
            perms = self.perms
            for k in perms:
                if k not in value:
                    self.removePermission(k)
            for k in value:
                self.addPermission(k)

    def __str__(self):
        return F"{self.group}:{self.member}:{self.id}"


class MembershipMetaData(MetaDataBase):
    parent = models.ForeignKey(Membership, related_name="properties", on_delete=models.CASCADE)



import collections
import inspect
import re
import sys

def guess_wrapper(cells):
    # look through the funciton closure for cells that look
    # like wrapped functions. for multuple functions,
    # attempt to guess which is the more likely candidate

    # we add points for
    # * non-lambda
    # * 'request' in arg list

    functions = collections.defaultdict(list)
    for cell in cells:
        contents = cell.cell_contents
        if not inspect.isfunction(contents):
            continue

        score = 0
        if contents.__name__ != '<lambda>':
            score += 1
        if 'request' in inspect.getargspec(contents).args:
            score += 1

        functions[score].append(contents)

    for score in reversed(sorted(functions.keys())):
        return functions[score][0]
    return None


def unwrap_and_compare(kwargs_provided, callback, description):
    # if the callback is decorated, unwrap to find the
    # original view

    # if we have a decorated function, the "wrapped"
    # function should b in the closure of the decorator
    # so we look for functions in the closure, guessing
    # if neccessary.

    if callback is not None:
        while callback.__closure__:
            compare(kwargs_provided, callback, description)
            cells = callback.__closure__
            guess = guess_wrapper(cells)
            if guess:
                callback = guess
            else:
                # nothing in the closure was a function, so give up
                break
        compare(kwargs_provided, callback, description)
        return callback



def compare(kwargs_provided, callback, description):
    spec = inspect.getargspec(callback)

    args = spec.args
    defaults = spec.defaults
    if defaults:
        required_args = args[:-len(defaults)]
    else:
        required_args = args

    missing_kwargs = set(required_args) - kwargs_provided
    if missing_kwargs:
        print(("%s: view requires kwargs %s not in the url kwargs" % (
            description, list(missing_kwargs))))


    if spec.keywords:
        # signature contains **kwargs, so can't do second check
        return

    extra_kwargs = kwargs_provided - set(args)
    if extra_kwargs:
        print(("%s: url provides kwargs %s not in the view signature" % (
            description, list(extra_kwargs))))



def check_resolver(entry, prefix):
    effective_regex = re.compile(prefix + entry.regex.pattern)

    # skip the admin
    if effective_regex.pattern.startswith('^admin/'):
        return
    #print dir(entry)
    description = (getattr(entry, 'name', None) or effective_regex.pattern)

    kwargs_provided = set(['request'] + list(getattr(entry, 'default_args', {}).keys()))
    #print kwargs_provided
    kwargs_provided.update(list(effective_regex.groupindex.keys()))
    callback = unwrap_and_compare(kwargs_provided, entry.callback, description)

    url_request = {"url":description, "methods":{}, "args": list(effective_regex.groupindex.keys())}
    if hasattr(entry, 'default_args'):
        dargs = getattr(entry, 'default_args', {})
        if "__MODULE" in dargs:
            url_request["module"] = dargs["__MODULE"].__name__
            method_patterns = dargs["__MODULE"].urlpattern_methods
            if entry.regex.pattern + "__GET" in method_patterns:
                url_request["methods"]["get"] = {"callback": method_patterns[entry.regex.pattern+ "__GET"]}
            if entry.regex.pattern + "__POST" in method_patterns:
                url_request["methods"]["post"] = {"callback": method_patterns[entry.regex.pattern+ "__POST"]}
            if entry.regex.pattern + "__DELETE" in method_patterns:
                url_request["methods"]["delete"] = {"callback": method_patterns[entry.regex.pattern+ "__DELETE"]}
            if entry.regex.pattern + "__PUT" in method_patterns:
                url_request["methods"]["put"] = {"callback": method_patterns[entry.regex.pattern+ "__PUT"]}
    if "module" not in url_request and callback:
        url_request["module"] = callback.__module__
    url_request["callback"] = callback


    #print dir(callback)
    return url_request


def show_urls(urllist, apis, filter=None, prefix=''):
    for entry in urllist:
        api = check_resolver(entry, prefix)
        if api:
            api["url"] = api["url"].replace("^", "").replace("$", "")
            if (filter and api["url"].startswith(filter)) or filter is None:
                if api["callback"].__doc__:
                    api["doc"] = api["callback"].__doc__
                else:
                    api["doc"] = None
                apis.append(api)

        if hasattr(entry, 'url_patterns'):
            show_urls(entry.url_patterns, apis, filter, prefix + entry.regex.pattern)

def sanatizeUrl(url):
    return re.sub(r'\(\?P<(.*?)>.*?\)',r'{\1}',url)

def parseDoc(api, graphs):
    """
    did regex, this approached seems to work better
    """
    doc = {"params":{}, "graphs":"", "summary":"", "issues":"", "returns":""}

    key = None
    subkey = None

    for line in api["doc"].split("\n"):
        linestr = line.strip()
        if linestr.startswith('|'):
            linestr = linestr[1:].strip()
            line = line.strip()[1:]
        lower_line = linestr.lower()
        if lower_line.startswith("parameter:") or lower_line.startswith("param:"):
            name = line[line.find(':')+1:line.find('=')].strip()
            info = line[line.find('=')+1:]
            doc["params"][name] = info
            key = "params"
            subkey = name
        elif lower_line.startswith("return:"):
            doc["returns"] = line[line.find(':')+1:].strip()
            key = "summary"
            subkey = None
        elif lower_line.startswith("issues:"):
            doc["issues"] = line[line.find(':')+1:].strip()
            key = "issues"
            subkey = None
        elif lower_line.startswith("graphs:"):
            doc["graphs"] = line[line.find(':')+1:].strip()
            key = "graphs"
            subkey = None
        elif key and subkey:
            doc[key][subkey] = "{0}\n{1}".format(doc[key][subkey], line.strip())
        elif key:
            doc[key] = "{0}\n{1}".format(doc[key], line.strip())
    doc["summary"] = doc["summary"].strip()

    if False: # doc["graphs"]
        graph_names = []
        fields = doc["graphs"].split('=')
        # print fields
        key = None
        value = None
        while len(fields):
            val = fields.pop(0)
            vals = val.strip()
            if vals.endswith("_graph"):
                epos = max(val.rfind(' '), val.rfind('\n'))
                nkey = val[epos:].strip()
                if epos > 0:
                    if value:
                        value = "{0}={1}".format(value, val[:epos].strip())
                    else:
                        value = val[:epos].strip()

                if key and value:
                    if key not in graphs:
                        graphs[key] = value
                    graph_names.append(key)
                key = nkey
            elif key:
                value = "{0}={1}".format(value, val)
        if key and value:
            if key not in graphs:
                graphs[key] = value
            graph_names.append(key)

        if "_graph" in doc["graphs"]:
            words = doc["graphs"].split()
            for word in words:
                if word.strip().endswith("_graph"):
                    graph_names.append(word.strip())
        doc["graphs"] = graph_names
    elif "_graph" in doc["returns"]:
        if "_graph" in doc["returns"]:
            graph_names = []
            words = doc["returns"].split()
            for word in words:
                if word.strip().endswith("_graph"):
                    graph_names.append(word.strip())
            doc["graphs"] = graph_names

    return doc

import json
def dumpGraphs(module_name, graphs):
    if module_name not in sys.modules:
        print("MODULE NOT LOADED.. loading")
        return
    module = sys.modules[module_name]
    for name in module.__dict__:
        # if "getMyself" in name:
        #     print "+++++++++++++++++"
        #     func = getattr(module, name)
        #     print func.__doc__
        #     print dir(func)
        #     print func.__dict__

        if "_graph" in name:
            graph = getattr(module, name)
            # print graph.keys()
            dc = {}
            if "fields" in graph:
                fields = graph["fields"]
                graph_doc = {}
                for f in fields:
                    if type(f) is tuple:
                        o = f[0]
                        f = f[1]
                        if len(f):
                            if "." in o:
                                o = o[:o.rfind('.')]
                                f = "{0}.{1}".format(o, f)
                    if "." in f:
                        levels = f.split('.')
                        current = graph_doc
                        field = levels[-1]
                        for o in levels:
                            if o is field:
                                continue
                            if o not in current:
                                current[o] = {}
                            elif type(current[o]) is str:
                                current[o] = {}
                            current = current[o]
                        current[field] = ''
                    else:
                        graph_doc[f] = ''
                graphs[name] = json.dumps(graph_doc)


def getRestApis(patterns, sanatize_urls=True):
    apis = []
    graphs = {}
    modules = []
    show_urls(patterns, apis)
    if sanatize_urls:
        old_apis = apis
        apis = []
        for api in old_apis:
            if len(api["methods"]):
                for method in api["methods"]:
                    a = {"method":method}
                    m = api["methods"][method]
                    module_name = m["callback"].__module__
                    if module_name not in modules:
                        dumpGraphs(module_name, graphs)
                        modules.append(module_name)

                    a["url"] = api["url"]
                    # print a["url"]

                    # # print m["callback"].func_doc
                    # if not m["callback"].__doc__:
                    #     print dir(m["callback"].__code__)
                    #     print m["callback"].__code__.__doc__
                    if m["callback"].__doc__:
                        # print "\tyes we have docs"
                        a["doc"] = m["callback"].__doc__
                        a["doc"] = parseDoc(a, graphs)
                        if a["doc"] is None or len(a["doc"]) == 0:
                            a["doc"] = ""
                        apis.append({
                            "module":api["module"].replace(".rpc", ""),
                            "method":method,
                            "url":sanatizeUrl(a["url"]),
                            "doc":a["doc"]
                        })
                    else:
                        apis.append({
                            "module":api["module"].replace(".rpc", ""),
                            "method":method,
                            "url":sanatizeUrl(a["url"]),
                            "doc":""
                        })
                        # print "\tno docs"
            else:
                # print api
                if api["doc"] is None or len(api["doc"]) == 0:
                    continue
                api["doc"] = parseDoc(api, graphs)
                apis.append({
                    "module":api["module"].replace(".rpc", ""),
                    "method":"*",
                    "url":sanatizeUrl(api["url"]),
                    "doc":api["doc"]
                    })
    return apis, graphs




# Generated by Django 4.1.4 on 2023-12-19 17:40

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('incident', '0012_rule_match_by'),
    ]

    operations = [
        migrations.AddField(
            model_name='rulecheck',
            name='is_required',
            field=models.IntegerField(default=0),
        ),
    ]




import base64
import binascii
import json
from io import BytesIO

try:
    from Crypto import Random
    from Crypto.PublicKey import RSA
    from Crypto.Cipher import AES, PKCS1_OAEP
except Exception:
    print("missing Crypto module.... pip install pycrypto")

from . import util


class PrivatePublicEncryption(object):
    def __init__(self, private_key=None, public_key=None, private_key_file=None, public_key_file=None):
        if private_key_file:
            with open(private_key_file, 'r') as f:
                private_key = f.read()
        if public_key_file:
            with open(public_key_file, 'r') as f:
                public_key = f.read()
        if private_key:
            if isinstance(private_key, str):
                private_key = RSA.importKey(private_key)
        self.private_key = private_key
        if public_key:
            if isinstance(public_key, str):
                public_key = RSA.importKey(public_key)
        self.public_key = public_key

    def generatePublicKey(self, make_new=False):
        if self.public_key is None or make_new:
            self.public_key = generatePublicKey(self.private_key)
        return self.public_key

    def encrypt(self, data):
        self.generatePublicKey()
        return self.encryptToB64(data)

    def decrypt(self, data, as_string=True):
        return self.decryptFromB64(data, as_string)

    def encryptToB64(self, data):
        # this function exports encrypted data as base64
        ebytes = encryptWithPublicKey(data, self.public_key)
        return util.toString(base64.b64encode(ebytes))

    def decryptFromB64(self, data, as_string=True):
        data = base64.b64decode(data)
        return decryptWithPrivateKey(data, self.private_key, as_string)

    def encryptToHex(self, data):
        # this function exports encrypted data as hex
        ebytes = encryptWithPublicKey(data, self.public_key)
        return util.toString(binascii.hexlify(ebytes))

    def decryptFromHex(self, data, as_string=True):
        data = binascii.unhexlify(data)
        return decryptWithPrivateKey(data, self.private_key, as_string)


def generatePrivateKey(size=2048):
    """
        key = RSA.generate(size)
        private_key = key.export_key()
        file_out = open("private.pem", "wb")
        file_out.write(private_key)
    """
    return RSA.generate(size)


def generatePublicKey(private_key):
    """
    public_key = private_key.publickey().export_key()
    file_out = open("receiver.pem", "wb")
    file_out.write(public_key)
    """
    if isinstance(private_key, str):
        private_key = RSA.import_key(private_key)
    return private_key.publickey()


def encryptWithPublicKey(data, public_key):
    if isinstance(public_key, str):
        public_key = RSA.import_key(public_key)

    if isinstance(data, dict) or isinstance(data, list):
        data = json.dumps(data)

    session_key = Random.get_random_bytes(16)

    # Encrypt the session key with the public RSA key
    cipher_rsa = PKCS1_OAEP.new(public_key)
    enc_session_key = cipher_rsa.encrypt(session_key)

    # Encrypt the data with the AES session key
    cipher_aes = AES.new(session_key, AES.MODE_EAX)
    ciphertext, tag = cipher_aes.encrypt_and_digest(util.toBytes(data))
    output = BytesIO()
    [output.write(x) for x in (enc_session_key, cipher_aes.nonce, tag, ciphertext)]
    raw = output.getvalue()
    output.close()
    return raw


def decryptWithPrivateKey(data, private_key, as_string=True):
    if isinstance(private_key, str):
        private_key = RSA.import_key(private_key)
    if isinstance(data, (str, bytes)):
        data = BytesIO(data)
    enc_session_key, nonce, tag, ciphertext = \
        [data.read(x) for x in (private_key.size_in_bytes(), 16, 16, -1)]

    cipher_rsa = PKCS1_OAEP.new(private_key)
    session_key = cipher_rsa.decrypt(enc_session_key)

    cipher_aes = AES.new(session_key, AES.MODE_EAX, nonce)
    decrypted_data = cipher_aes.decrypt_and_verify(ciphertext, tag)
    if as_string:
        return util.toString(decrypted_data)
    return decrypted_data



from django.http import HttpResponse

from rest import decorators as rd
from medialib.qrcode import generateQRCode


@rd.url(r'^qrcode$')
def qrcode_image(request):
    params = dict(data=request.DATA.get("data", "missing data"))
    error = request.DATA.get("error", None)
    if error is not None:
        params["error"] = error
    version = request.DATA.get("version", None)
    if version is not None:
        params["version"] = int(version)
    img_format = request.DATA.get("format", "png")
    if img_format is not None:
        params["img_format"] = img_format
    scale = request.DATA.get("scale", 1)
    if scale is not None:
        params["scale"] = int(scale)

    code = generateQRCode(**params)
    if img_format == "base64":
        return HttpResponse(code, content_type="text/plain")
    elif img_format == "svg":
        return HttpResponse(code, content_type="image/svg+xml")
    return HttpResponse(code, content_type="image/png")



# Generated by Django 4.1.4 on 2023-12-25 20:46

from django.db import migrations, models
import django.db.models.deletion
import rest.fields
import rest.models.base


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0017_rename_requires_topt_member_requires_totp'),
    ]

    operations = [
        migrations.CreateModel(
            name='UserPassKey',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('uuid', models.CharField(max_length=255, unique=True)),
                ('rp_id', models.CharField(db_index=True, max_length=255)),
                ('is_enabled', models.BooleanField(db_index=True, default=True)),
                ('name', models.CharField(max_length=255)),
                ('platform', models.CharField(default='', max_length=255)),
                ('last_used', models.DateTimeField(default=None, null=True)),
                ('token', models.CharField(max_length=255)),
                ('info', rest.fields.JSONField(default=None, null=True)),
                ('member', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='passkeys', to='account.member')),
            ],
            bases=(models.Model, rest.models.base.RestModel),
        ),
    ]



# Early, and incomplete implementation of -04.
#
import re
import urllib.request, urllib.parse, urllib.error

RESERVED = ":/?#[]@!$&'()*+,;="
OPERATOR = "+./;?|!@"
EXPLODE = "*+"
MODIFIER = ":^"
TEMPLATE = re.compile(r"{(?P<operator>[\+\./;\?|!@])?(?P<varlist>[^}]+)}", re.UNICODE)
VAR = re.compile(r"^(?P<varname>[^=\+\*:\^]+)((?P<explode>[\+\*])|(?P<partial>[:\^]-?[0-9]+))?(=(?P<default>.*))?$", re.UNICODE)

def _tostring(varname, value, explode, operator, safe=""):
  if type(value) == type([]):
    if explode == "+":
      return ",".join([varname + "." + urllib.parse.quote(x, safe) for x in value])
    else:
      return ",".join([urllib.parse.quote(x, safe) for x in value])
  if type(value) == type({}):
    keys = list(value.keys())
    keys.sort()
    if explode == "+":
      return ",".join([varname + "." + urllib.parse.quote(key, safe) + "," + urllib.parse.quote(value[key], safe) for key in keys])
    else:
      return ",".join([urllib.parse.quote(key, safe) + "," + urllib.parse.quote(value[key], safe) for key in keys])
  else:
    return urllib.parse.quote(value, safe)


def _tostring_path(varname, value, explode, operator, safe=""):
  joiner = operator
  if type(value) == type([]):
    if explode == "+":
      return joiner.join([varname + "." + urllib.parse.quote(x, safe) for x in value])
    elif explode == "*":
      return joiner.join([urllib.parse.quote(x, safe) for x in value])
    else:
      return ",".join([urllib.parse.quote(x, safe) for x in value])
  elif type(value) == type({}):
    keys = list(value.keys())
    keys.sort()
    if explode == "+":
      return joiner.join([varname + "." + urllib.parse.quote(key, safe) + joiner + urllib.parse.quote(value[key], safe) for key in keys])
    elif explode == "*":
      return joiner.join([urllib.parse.quote(key, safe) + joiner + urllib.parse.quote(value[key], safe) for key in keys])
    else:
      return ",".join([urllib.parse.quote(key, safe) + "," + urllib.parse.quote(value[key], safe) for key in keys])
  else:
    if value:
      return urllib.parse.quote(value, safe)
    else:
      return ""

def _tostring_query(varname, value, explode, operator, safe=""):
  joiner = operator
  varprefix = ""
  if operator == "?":
    joiner = "&"
    varprefix = varname + "="
  if type(value) == type([]):
    if 0 == len(value):
      return ""
    if explode == "+":
      return joiner.join([varname + "=" + urllib.parse.quote(x, safe) for x in value])
    elif explode == "*":
      return joiner.join([urllib.parse.quote(x, safe) for x in value])
    else:
      return varprefix + ",".join([urllib.parse.quote(x, safe) for x in value])
  elif type(value) == type({}):
    if 0 == len(value):
      return ""
    keys = list(value.keys())
    keys.sort()
    if explode == "+":
      return joiner.join([varname + "." + urllib.parse.quote(key, safe) + "=" + urllib.parse.quote(value[key], safe) for key in keys])
    elif explode == "*":
      return joiner.join([urllib.parse.quote(key, safe) + "=" + urllib.parse.quote(value[key], safe) for key in keys])
    else:
      return varprefix + ",".join([urllib.parse.quote(key, safe) + "," + urllib.parse.quote(value[key], safe) for key in keys])
  else:
    if value:
      return varname + "=" + urllib.parse.quote(value, safe)
    else:
      return varname 

TOSTRING = {
    "" : _tostring,
    "+": _tostring,
    ";": _tostring_query,
    "?": _tostring_query,
    "/": _tostring_path,
    ".": _tostring_path,
    }


def expand(template, vars):
  def _sub(match):
    groupdict = match.groupdict()
    operator = groupdict.get('operator')
    if operator is None:
      operator = ''
    varlist = groupdict.get('varlist')

    safe = "@"
    if operator == '+':
      safe = RESERVED
    varspecs = varlist.split(",")
    varnames = []
    defaults = {}
    for varspec in varspecs:
      m = VAR.search(varspec)
      groupdict = m.groupdict()
      varname = groupdict.get('varname')
      explode = groupdict.get('explode')
      partial = groupdict.get('partial')
      default = groupdict.get('default')
      if default:
        defaults[varname] = default
      varnames.append((varname, explode, partial))

    retval = []
    joiner = operator
    prefix = operator
    if operator == "+":
      prefix = ""
      joiner = ","
    if operator == "?":
      joiner = "&"
    if operator == "":
      joiner = ","
    for varname, explode, partial in varnames:
      if varname in vars:
        value = vars[varname]
        #if not value and (type(value) == type({}) or type(value) == type([])) and varname in defaults:
        if not value and value != "" and varname in defaults:
          value = defaults[varname]
      elif varname in defaults:
        value = defaults[varname]
      else:
        continue
      retval.append(TOSTRING[operator](varname, value, explode, operator, safe=safe))
    if "".join(retval):
      return prefix + joiner.join(retval)
    else:
      return ""

  return TEMPLATE.sub(_sub, template)



from datetime import datetime
from rest import decorators as rd
from rest import views as rv
from rest import settings
from rest import helpers as rh
from rest import net
from rest import ssl_check
from rest import helpers as rh
from location.models import GeoIP
from . import client as metrics
from . import models as mm
from . import models as em
from .providers import aws
from objict import nobjict, objict
import time
import re

"""
Capture simple analytics counters.
"""
LOCATION_METRICS = settings.LOCATION_METRICS
REST_PREFIX = settings.get("REST_PREFIX", "api/")


@rd.urlPOST('metric')
def rest_on_new_metric(request):
    # slug, num=1, category=None, expire=None, date=None
    data = request.DATA.toObject()
    if data.slug is None:
        return rv.restStatus(request, False)
    if data.value is None:
        data.value = 1
    metrics.metric(data.slug, data.value, category=data.category, expire=data.expire)
    if request.DATA.get("geolocate", False, field_type=bool):
        if request.location is None:
            request.location = GeoIP.get(request.ip)
        if request.location and request.location.country:
            country = request.location.country.lower().replace(" ", "_")
            state = request.location.state.lower().replace(" ", "_")
            metrics.metric(f"geo_country__{country}__{data.slug}", category="geo_country")
            metrics.metric(f"geo_state__{state}__{data.slug}", category="geo_state")
            
    return rv.restStatus(request, True)


@rd.urlGET('metric')
@rd.login_required
def rest_on_get_metric(request):
    # slug, num=1, category=None, expire=None, date=None
    data = request.DATA.toObject()
    if data.slug is None:
        return rv.restStatus(request, False)
    result = metrics.get_metric(data.slug)
    if result is None:
        return rv.restStatus(request, False)
    return rv.restReturn(request, dict(data=result))


@rd.urlGET('metrics')
@rd.login_required
def rest_on_get_metrics(request, pk=None):
    # slug, since, granularity
    since = request.DATA.get("since", field_type=datetime)
    granularity = request.DATA.get(["granularity", "period"], default="daily")
    samples = request.DATA.get("samples", field_type=int)
    category = request.DATA.get("category")
    rh.debug(f"rest_on_get_metrics: {since}, {granularity}, {samples}, {category}")
    if category:
        result = metrics.get_category_metrics(category, since, granularity, samples=samples)
        return rv.restReturn(request, dict(data=result))
    slugs = request.DATA.getlist(["slugs", "slug"])
    if slugs is None:
        return rv.restStatus(request, False)

    result = metrics.get_metrics(slugs, since, granularity, samples=samples)
    if result is None:
        return rv.restStatus(request, False)
    return rv.restReturn(request, dict(data=result))


@rd.urlGET('slugs')
@rd.login_required
def rest_on_get_metrics_slugs(request, pk=None):
    category = request.DATA.get("category", None)
    slugs = metrics.get_slugs(category)
    prefix = request.DATA.get("prefix")
    if prefix:
        slugs = [s for s in slugs if s.startswith(prefix)]
    return rv.restReturn(request, dict(data=slugs))


@rd.urlPOST('guage')
def rest_on_guage(request, pk=None):
    data = request.DATA.toObject()
    if data.slug is None or data.value is None:
        return rv.restStatus(request, False)
    metrics.guage(data.slug, data.value)
    return rv.restStatus(request, True)


@rd.urlGET('guage')
@rd.login_required
def rest_on_get_guage(request):
    # slug, num=1, category=None, expire=None, date=None
    data = request.DATA.toObject()
    if data.slug is None:
        return rv.restStatus(request, False)
    return rv.restReturn(request, metrics.get_guage(data.slug))


@rd.urlGET('guages')
@rd.login_required
def rest_on_get_guages(request, pk=None):
    # slug, num=1, category=None, expire=None, date=None
    data = request.DATA.toObject()
    if data.slugs is None:
        return rv.restStatus(request, False)
    return rv.restReturn(request, metrics.get_guages(data.slugs))


@rd.url('db/metrics/objects')
@rd.url('db/metrics/objects/<int:pk>')
@rd.login_required
def rest_on_db_metrics_objects(request, pk=None):
    return mm.Metrics.on_rest_request(request, pk)


@rd.urlGET('db/metrics')
@rd.login_required
def rest_on_get_model_metrics(request, pk=None):
    # slug, since, granularity
    since = request.DATA.get("since", field_type=datetime)
    granularity = request.DATA.get(["granularity", "period"], default="daily")
    slugs = request.DATA.getlist(["slug", "slugs"])
    if not slugs:
        return rv.restStatus(request, False)
    status, reason, code = rh.requestHasPerms(request, ["view_metrics", "reporting", "admin", "manage_groups"], group=request.group)
    if not status:
        return rv.restPermissionDenied(request, reason, code)
    to = None
    if "to" in request.DATA:
        to = request.DATA.get("to", field_type=datetime)
    field = request.DATA.get("field")
    if field is not None:
        return on_get_metrics_by_field(request, slugs, field, since, to, granularity)

    result = mm.get_metrics(slugs[0], granularity, since, group=request.group)
    if result is None:
        return rv.restStatus(request, False)
    return rv.restReturn(request, dict(data=result))


def on_get_metrics_by_field(request, slugs, field, since, to, granularity):
    output = nobjict(periods=None, data=nobjict())
    samples = request.DATA.get("samples")
    field2 = None
    if "__" in field:
        field, field2 = field.split("__")
    for slug in slugs:
        result = mm.get_metrics(slug, granularity, since, to, samples=samples)
        if result is None:
            continue
        if output.periods is None and result.periods:
            output.periods = result.periods
        data_set = result.data.get(field, None)
        if data_set is None:
            output.error = f"missing field:{field} from data {result.data}"
        elif field2:
            d2 = result.data.get(field2, None)
            d1 = data_set
            data_set = [int(d1[i]*100/d2[i]) if d1[i] > 0 and d2[i] > 0 else 0 for i in range(0, len(d1))]
        output.data[slug] = data_set

    # attempt to fix data set issues
    if not output.periods:
        output.periods = mm.get_chart_periods(slug, granularity, since)
    for key in output.data:
        if output.data[key] is None:
            output.data[key] = [0 for i in range(0, len(output.periods))]
    return rv.restReturn(request, dict(data=output))


@rd.urlGET('db/metric')
@rd.login_required
def rest_on_get_model_metric(request, pk=None):
    # slug, since, granularity
    since = request.DATA.get("since", field_type=datetime)
    granularity = request.DATA.get(["granularity", "period"], default="daily")
    slugs = request.DATA.get(["slug", "slugs"])
    if slugs is None:
        return rv.restStatus(request, False)

    status, reason, code = rh.requestHasPerms(request, ["view_metrics", "reporting", "admin", "manage_groups"], group=request.group)
    if not status:
        return rv.restPermissionDenied(request, reason, code)

    result = mm.get_metric(slugs, granularity, since, group=request.group)
    if result is None:
        return rv.restStatus(request, False)
    return rv.restReturn(request, dict(data=result))


@rd.urlGET('db/slugs')
@rd.login_required
def rest_on_get_model_metrics_slugs(request, pk=None):
    return rv.restGet(request, dict(data=list(mm.Metrics.objects.all().distinct().values_list("slug", flat=True))))


# BEGIN EOD METRICS
@rd.urlGET('db/eod/metric')
@rd.login_required
def rest_on_get_model_eod_metrics(request, pk=None):
    # slug, since, granularity
    dr_end = request.DATA.get("dr_end", field_type=datetime)
    slug = request.DATA.get("slug")
    status, reason, code = rh.requestHasPerms(request, ["view_metrics", "reporting", "admin", "manage_groups"], group=request.group)
    if not status:
        return rv.restPermissionDenied(request, reason, code)

    result = em.get_metric(slug, dr_end, group=request.group)
    if result is None:
        return rv.restStatus(request, False)
    return rv.restReturn(request, dict(data=result))


@rd.urlGET('db/eod/metrics')
@rd.login_required
def rest_on_get_model_eod_metrics(request, pk=None):
    # slug, since, granularity
    dr_end = request.DATA.get("dr_end", field_type=datetime)
    slug = request.DATA.get("slug")
    count = request.DATA.get("count", 7)
    status, reason, code = rh.requestHasPerms(request, ["view_metrics", "reporting", "admin", "manage_groups"], group=request.group)
    if not status:
        return rv.restPermissionDenied(request, reason, code)

    result = em.get_metrics(slug, count, dr_end, group=request.group)
    if result is None:
        return rv.restStatus(request, False)
    return rv.restReturn(request, dict(data=result))


@rd.url('db/eod/metrics/objects')
@rd.url('db/eod/metrics/objects/<int:pk>')
@rd.login_required
def rest_on_db_eod_metrics_objects(request, pk=None):
    return em.EODMetrics.on_rest_request(request, pk)


@rd.urlGET('db/eod/slugs')
@rd.login_required
def rest_on_get_model_eod_metrics_slugs(request, pk=None):
    return rv.restGet(request, dict(data=list(em.EODMetrics.objects.all().distinct().values_list("slug", flat=True))))


# BEGIN CLOUDWATCH

def getPeriods(period, duration):
    count = int(duration/period)
    periods = []
    utc_time = int(time.strftime('%s', time.gmtime()))
    for i in range(0, count-1):
        periods.append(utc_time - (i * period))
    periods.reverse()
    return periods


@rd.urlGET('restit/domains')
@rd.login_required
def rest_on_ec2_domains(request, pk=None):
    if not settings.DOMAIN_WATCH:
        return rv.restPermissionDenied(request)

    output = []
    result = ssl_check.check(*settings.DOMAIN_WATCH)
    for key, value in result.items():
        output.append(dict(domain=key, expires=value, id=key))
    return rv.restList(request, output)


@rd.urlGET('restit/servers')
@rd.login_required
def rest_on_ec2_restit_stats(request, pk=None):
    hostname = settings.SERVER_ROOT
    if hostname is None:
        hostname = settings.SERVER_NAME
    if hostname.count(".") == 2:
        hostname = hostname[hostname.find(".")]
    instances = aws.getAllEC2()
    hosts = [inst.name for inst in instances]
    data = []
    for name in hosts:
        if settings.SERVER_NAME_MAP and name in settings.SERVER_NAME_MAP:
            host = settings.SERVER_NAME_MAP[name]
            if "." not in host:
                host = f"{name}.{hostname}"
        else:
            host = f"{name}.{hostname}"
        resp = net.REQUEST("GET", host, f"/{REST_PREFIX}versions", params=dict(detailed=1))
        if resp.status:
            resp.data.id = host
            resp.data.hostname = host
            data.append(resp.data)
    return rv.restReturn(request, dict(data=data))


@rd.urlGET('aws/ec2/list/details')
@rd.login_required
def rest_on_ec2_list(request, pk=None):
    return rv.restReturn(request, dict(data=aws.getAllEC2()))


@rd.urlGET('aws/ec2/list/ids')
@rd.login_required
def rest_on_ec2_list_ids(request, pk=None):
    return rv.restReturn(request, dict(data=aws.getAllEC2(True)))


@rd.urlGET('aws/ec2/cpu')
@rd.requires_params(["ids"])
def rest_on_ec2_metrics_cpu(request, pk=None):
    ids = request.DATA.getlist("ids")
    period = request.DATA.get("period", 300, field_type=int)
    duration = request.DATA.get("duration", period*8, field_type=int)
    get_rds = False
    instances = aws.getAllEC2()
    if "all" in ids:
        get_rds = True
        ids = [inst.id for inst in instances]
    data = aws.getMetrics(
        ids,
        period=period,
        duration_seconds=duration,
        metric="cpu",
        namespace="ec2",
        stat="max")
    # convert ids to names
    output = dict()
    for i in instances:
        if i.id in data:
            output[i.name] = data[i.id]
            output[i.name].reverse()
    # if get_rds:
    #     ids = aws.getAllRDS(True)
    #     rds_cpu = (aws.getMetrics(
    #         ids,
    #         period=period,
    #         duration_seconds=duration,
    #         metric="cpu",
    #         namespace="rds",
    #         stat="max"))
    #     for k, v in rds_cpu.items():
    #         v.reverse()
    #     output.update(rds_cpu)

    data = dict(data=output, periods=getPeriods(period, duration))
    return rv.restReturn(request, dict(data=data))


@rd.urlGET('aws/ec2/network/in')
@rd.requires_params(["ids"])
def rest_on_ec2_metrics_netin(request, pk=None):
    data = aws.getMetrics(
        request.DATA.getlist("ids"),
        period=request.DATA.get("period", 300, field_type=int),
        duration_seconds=request.DATA.get("duration", 900, field_type=int),
        metric="NetworkIn",
        namespace="ec2",
        stat="max")
    return rv.restReturn(request, dict(data=data))


@rd.urlGET('aws/ec2/network/out')
@rd.requires_params(["ids"])
def rest_on_ec2_metrics_netout(request, pk=None):
    ids = request.DATA.getlist("ids")
    period = request.DATA.get("period", 300, field_type=int)
    duration = request.DATA.get("duration", period*8, field_type=int)
    get_rds = False
    instances = aws.getAllEC2()
    if "all" in ids:
        get_rds = True
        ids = [inst.id for inst in instances]
    data = aws.getMetrics(
        ids,
        period=period,
        duration_seconds=duration,
        metric="NetworkOut",
        namespace="ec2",
        stat="max")
    # convert ids to names
    output = dict()
    for i in instances:
        if i.id in data:
            output[i.name] = data[i.id]
            output[i.name].reverse()
    data = dict(data=output, periods=getPeriods(period, duration))
    return rv.restReturn(request, dict(data=data))


@rd.urlGET('aws/rds/cpu')
@rd.requires_params(["ids"])
def rest_on_ec2_metrics_cpu(request, pk=None):
    ids = request.DATA.getlist("ids")
    period = request.DATA.get("period", 300, field_type=int)
    duration = request.DATA.get("duration", period*8, field_type=int)
    instances = aws.getAllRDS()
    if "all" in ids:
        ids = [inst.id for inst in instances]
    data = aws.getMetrics(
        ids,
        period=period,
        duration_seconds=duration,
        metric="cpu",
        namespace="rds",
        stat="max")
    data = dict(data=data, periods=getPeriods(period, duration))
    return rv.restReturn(request, dict(data=data))


@rd.urlGET('aws/rds/memory')
@rd.requires_params(["ids"])
def rest_on_rds_metrics_memory(request, pk=None):
    ids = request.DATA.getlist("ids")
    period = request.DATA.get("period", 300, field_type=int)
    duration = request.DATA.get("duration", period*8, field_type=int)
    instances = aws.getAllRDS()
    if "all" in ids:
        ids = [inst.id for inst in instances]
    data = aws.getMetrics(
        ids,
        period=period,
        duration_seconds=duration,
        metric="memory",
        namespace="rds",
        stat="max")
    data = dict(data=data, periods=getPeriods(period, duration))
    return rv.restReturn(request, dict(data=data))


@rd.urlGET('aws/rds/cons')
@rd.requires_params(["ids"])
def rest_on_rds_metrics_cons(request, pk=None):
    ids = request.DATA.getlist("ids")
    period = request.DATA.get("period", 300, field_type=int)
    duration = request.DATA.get("duration", period*8, field_type=int)
    instances = aws.getAllRDS()
    if "all" in ids:
        ids = [inst.id for inst in instances]
    data = aws.getMetrics(
        ids,
        period=period,
        duration_seconds=duration,
        metric="cons",
        namespace="rds",
        stat="max")
    data = dict(data=data, periods=getPeriods(period, duration))
    return rv.restReturn(request, dict(data=data))


@rd.urlGET('aws/redis/cpu')
@rd.requires_params(["ids"])
def rest_on_redis_metrics_cpu(request, pk=None):
    return get_redis_metrics(request, "cpu")


@rd.urlGET('aws/redis/cons')
@rd.requires_params(["ids"])
def rest_on_redis_metrics_cons(request, pk=None):
    return get_redis_metrics(request, "cache_cons")


@rd.urlGET('aws/redis/memory')
@rd.requires_params(["ids"])
def rest_on_redis_metrics_mem(request, pk=None):
    return get_redis_metrics(request, "cache_usage")


def get_redis_metrics(request, metric):
    ids = request.DATA.getlist("ids")
    period = request.DATA.get("period", 300, field_type=int)
    duration = request.DATA.get("duration", period*8, field_type=int)
    instances = aws.getAllRedis()
    if "all" in ids:
        ids = [inst.id for inst in instances]
    data = aws.getMetrics(
        ids,
        period=period,
        duration_seconds=duration,
        metric=metric,
        namespace="redis",
        stat="max")
    data = dict(data=data, periods=getPeriods(period, duration))
    return rv.restReturn(request, dict(data=data))


# @rd.urlGET('aws/ec2/metrics/list')
# @rd.login_required
# def rest_on_ec2_list_ids(request, pk=None):
#     ['DiskWriteOps', 'CPUCreditUsage', 'DiskReadOps', 
#     'NetworkOut', 'CPUSurplusCreditBalance', 
#     'CPUUtilization', 'CPUCreditBalance', 
#     'NetworkPacketsOut', 'StatusCheckFailed', 
#     'NetworkIn', 'StatusCheckFailed_System', 
#     'StatusCheckFailed_Instance', 'DiskWriteBytes', 
#     'NetworkPacketsIn', 'CPUSurplusCreditsCharged', 
#     'DiskReadBytes', 'MetadataNoToken']
#     return rv.restReturn(request, data=aws.getMetricsList(request.DATA.get("id", None)))


@rd.urlGET('aws/rds/list/details')
@rd.login_required
def rest_on_rds_list(request, pk=None):
    return rv.restReturn(request, dict(data=aws.getAllRDS()))


@rd.urlGET('aws/rds/list/ids')
@rd.login_required
def rest_on_rds_list_ids(request, pk=None):
    return rv.restReturn(request, dict(data=aws.getAllRDS(True)))


@rd.urlGET('logs/nginx')
@rd.perm_required("view_logs")
def rest_on_get_logs(request):
    ip = request.DATA.get("ip", None)
    gz = request.DATA.get("gz", None)
    if ip is None:
        return rv.restPermissionDenied(request)
    return rv.restReturn(request, dict(data=search_logs(ip, gz)))


def search_logs(ip, gz, custom=None):
    cmd = ["/opt/api/bin/logsearch.sh", ip]
    if gz:
        cmd.append(gz)
    out, err = rh.sudoCMD(cmd, return_output=True)
    output = []
    index = 0
    for line in out.decode().split("\n"):
        jline = nginx_line_to_json(line)
        if jline is not None:
            index += 1
            jline["id"] = f"{settings.SERVER_NAME}{index}"
            output.append(jline)
    return output


NGINX_PATTERN = r'(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}) (.*) (.*) \[(.*?)\] "(.*?)" (\d{3}) (\d*) "(.*?)" "(.*?)"'
def nginx_line_to_json(line):
    prog = re.compile(NGINX_PATTERN)
    match = prog.match(line)
    if match is None:
        return None
    return {
        'server': settings.SERVER_NAME,
        'remote_addr': match.group(1),
        'sent_http_x_sessionid': match.group(2),
        'sent_http_x_uid': match.group(3),
        'time_local': match.group(4),
        'request': match.group(5),
        'status': int(match.group(6)),
        'body_bytes_sent': int(match.group(7)),
        'referrer': match.group(8),
        'http_user_agent': match.group(9)
    }


@rd.urlGET('servers/logs')
@rd.perm_required("view_logs")
def rest_on_ec2_logs(request, pk=None):
    hostname = settings.SERVER_ROOT
    if hostname is None:
        hostname = settings.SERVER_NAME
    if hostname.count(".") == 2:
        hostname = hostname[hostname.find(".")]
    instances = aws.getAllEC2()
    hosts = [inst.name for inst in instances]
    ip = request.DATA.get("ip", "headless")
    gz = request.DATA.get("gz", None)
    params = dict(ip=ip, gz=gz)
    logs = []
    auth = request.META.get("HTTP_AUTHORIZATION", None)
    if not auth or not ip:
        return rv.restPermissionDenied(request)
    headers = dict(Authorization=auth, referer=settings.BASE_URL)
    for name in hosts:
        if settings.SERVER_NAME_MAP and name in settings.SERVER_NAME_MAP:
            host = settings.SERVER_NAME_MAP[name]
            if "." not in host:
                host = f"{name}.{hostname}"
        else:
            host = f"{name}.{hostname}"
        resp = net.REQUEST(
            "GET", host,
            f"/{REST_PREFIX}metrics/logs/nginx",
            headers=headers, params=params)
        if resp.status and resp.data:
            logs.extend(resp.data)
    return rv.restReturn(request, dict(data=logs, size=len(logs), count=len(logs)))




# Copyright 2011 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Locked file interface that should work on Unix and Windows pythons.

This module first tries to use fcntl locking to ensure serialized access
to a file, then falls back on a lock file if that is unavialable.

Usage:
    f = LockedFile('filename', 'r+b', 'rb')
    f.open_and_lock()
    if f.is_locked():
      print 'Acquired filename with r+b mode'
      f.file_handle().write('locked data')
    else:
      print 'Aquired filename with rb mode'
    f.unlock_and_close()
"""

__author__ = 'cache@google.com (David T McWherter)'

import errno
import logging
import os
import time

from oauth2client import util

logger = logging.getLogger(__name__)


class CredentialsFileSymbolicLinkError(Exception):
  """Credentials files must not be symbolic links."""


class AlreadyLockedException(Exception):
  """Trying to lock a file that has already been locked by the LockedFile."""
  pass


def validate_file(filename):
  if os.path.islink(filename):
    raise CredentialsFileSymbolicLinkError(
        'File: %s is a symbolic link.' % filename)

class _Opener(object):
  """Base class for different locking primitives."""

  def __init__(self, filename, mode, fallback_mode):
    """Create an Opener.

    Args:
      filename: string, The pathname of the file.
      mode: string, The preferred mode to access the file with.
      fallback_mode: string, The mode to use if locking fails.
    """
    self._locked = False
    self._filename = filename
    self._mode = mode
    self._fallback_mode = fallback_mode
    self._fh = None

  def is_locked(self):
    """Was the file locked."""
    return self._locked

  def file_handle(self):
    """The file handle to the file. Valid only after opened."""
    return self._fh

  def filename(self):
    """The filename that is being locked."""
    return self._filename

  def open_and_lock(self, timeout, delay):
    """Open the file and lock it.

    Args:
      timeout: float, How long to try to lock for.
      delay: float, How long to wait between retries.
    """
    pass

  def unlock_and_close(self):
    """Unlock and close the file."""
    pass


class _PosixOpener(_Opener):
  """Lock files using Posix advisory lock files."""

  def open_and_lock(self, timeout, delay):
    """Open the file and lock it.

    Tries to create a .lock file next to the file we're trying to open.

    Args:
      timeout: float, How long to try to lock for.
      delay: float, How long to wait between retries.

    Raises:
      AlreadyLockedException: if the lock is already acquired.
      IOError: if the open fails.
      CredentialsFileSymbolicLinkError if the file is a symbolic link.
    """
    if self._locked:
      raise AlreadyLockedException('File %s is already locked' %
                                   self._filename)
    self._locked = False

    validate_file(self._filename)
    try:
      self._fh = open(self._filename, self._mode)
    except IOError as e:
      # If we can't access with _mode, try _fallback_mode and don't lock.
      if e.errno == errno.EACCES:
        self._fh = open(self._filename, self._fallback_mode)
        return

    lock_filename = self._posix_lockfile(self._filename)
    start_time = time.time()
    while True:
      try:
        self._lock_fd = os.open(lock_filename,
                                os.O_CREAT|os.O_EXCL|os.O_RDWR)
        self._locked = True
        break

      except OSError as e:
        if e.errno != errno.EEXIST:
          raise
        if (time.time() - start_time) >= timeout:
          logger.warn('Could not acquire lock %s in %s seconds' % (
              lock_filename, timeout))
          # Close the file and open in fallback_mode.
          if self._fh:
            self._fh.close()
          self._fh = open(self._filename, self._fallback_mode)
          return
        time.sleep(delay)

  def unlock_and_close(self):
    """Unlock a file by removing the .lock file, and close the handle."""
    if self._locked:
      lock_filename = self._posix_lockfile(self._filename)
      os.close(self._lock_fd)
      os.unlink(lock_filename)
      self._locked = False
      self._lock_fd = None
    if self._fh:
      self._fh.close()

  def _posix_lockfile(self, filename):
    """The name of the lock file to use for posix locking."""
    return '%s.lock' % filename


try:
  import fcntl

  class _FcntlOpener(_Opener):
    """Open, lock, and unlock a file using fcntl.lockf."""

    def open_and_lock(self, timeout, delay):
      """Open the file and lock it.

      Args:
        timeout: float, How long to try to lock for.
        delay: float, How long to wait between retries

      Raises:
        AlreadyLockedException: if the lock is already acquired.
        IOError: if the open fails.
        CredentialsFileSymbolicLinkError if the file is a symbolic link.
      """
      if self._locked:
        raise AlreadyLockedException('File %s is already locked' %
                                     self._filename)
      start_time = time.time()

      validate_file(self._filename)
      try:
        self._fh = open(self._filename, self._mode)
      except IOError as e:
        # If we can't access with _mode, try _fallback_mode and don't lock.
        if e.errno == errno.EACCES:
          self._fh = open(self._filename, self._fallback_mode)
          return

      # We opened in _mode, try to lock the file.
      while True:
        try:
          fcntl.lockf(self._fh.fileno(), fcntl.LOCK_EX)
          self._locked = True
          return
        except IOError as e:
          # If not retrying, then just pass on the error.
          if timeout == 0:
            raise e
          if e.errno != errno.EACCES:
            raise e
          # We could not acquire the lock. Try again.
          if (time.time() - start_time) >= timeout:
            logger.warn('Could not lock %s in %s seconds' % (
                self._filename, timeout))
            if self._fh:
              self._fh.close()
            self._fh = open(self._filename, self._fallback_mode)
            return
          time.sleep(delay)

    def unlock_and_close(self):
      """Close and unlock the file using the fcntl.lockf primitive."""
      if self._locked:
        fcntl.lockf(self._fh.fileno(), fcntl.LOCK_UN)
      self._locked = False
      if self._fh:
        self._fh.close()
except ImportError:
  _FcntlOpener = None


try:
  import pywintypes
  import win32con
  import win32file

  class _Win32Opener(_Opener):
    """Open, lock, and unlock a file using windows primitives."""

    # Error #33:
    #  'The process cannot access the file because another process'
    FILE_IN_USE_ERROR = 33

    # Error #158:
    #  'The segment is already unlocked.'
    FILE_ALREADY_UNLOCKED_ERROR = 158

    def open_and_lock(self, timeout, delay):
      """Open the file and lock it.

      Args:
        timeout: float, How long to try to lock for.
        delay: float, How long to wait between retries

      Raises:
        AlreadyLockedException: if the lock is already acquired.
        IOError: if the open fails.
        CredentialsFileSymbolicLinkError if the file is a symbolic link.
      """
      if self._locked:
        raise AlreadyLockedException('File %s is already locked' %
                                     self._filename)
      start_time = time.time()

      validate_file(self._filename)
      try:
        self._fh = open(self._filename, self._mode)
      except IOError as e:
        # If we can't access with _mode, try _fallback_mode and don't lock.
        if e.errno == errno.EACCES:
          self._fh = open(self._filename, self._fallback_mode)
          return

      # We opened in _mode, try to lock the file.
      while True:
        try:
          hfile = win32file._get_osfhandle(self._fh.fileno())
          win32file.LockFileEx(
              hfile,
              (win32con.LOCKFILE_FAIL_IMMEDIATELY|
               win32con.LOCKFILE_EXCLUSIVE_LOCK), 0, -0x10000,
              pywintypes.OVERLAPPED())
          self._locked = True
          return
        except pywintypes.error as e:
          if timeout == 0:
            raise e

          # If the error is not that the file is already in use, raise.
          if e[0] != _Win32Opener.FILE_IN_USE_ERROR:
            raise

          # We could not acquire the lock. Try again.
          if (time.time() - start_time) >= timeout:
            logger.warn('Could not lock %s in %s seconds' % (
                self._filename, timeout))
            if self._fh:
              self._fh.close()
            self._fh = open(self._filename, self._fallback_mode)
            return
          time.sleep(delay)

    def unlock_and_close(self):
      """Close and unlock the file using the win32 primitive."""
      if self._locked:
        try:
          hfile = win32file._get_osfhandle(self._fh.fileno())
          win32file.UnlockFileEx(hfile, 0, -0x10000, pywintypes.OVERLAPPED())
        except pywintypes.error as e:
          if e[0] != _Win32Opener.FILE_ALREADY_UNLOCKED_ERROR:
            raise
      self._locked = False
      if self._fh:
        self._fh.close()
except ImportError:
  _Win32Opener = None


class LockedFile(object):
  """Represent a file that has exclusive access."""

  @util.positional(4)
  def __init__(self, filename, mode, fallback_mode, use_native_locking=True):
    """Construct a LockedFile.

    Args:
      filename: string, The path of the file to open.
      mode: string, The mode to try to open the file with.
      fallback_mode: string, The mode to use if locking fails.
      use_native_locking: bool, Whether or not fcntl/win32 locking is used.
    """
    opener = None
    if not opener and use_native_locking:
      if _Win32Opener:
        opener = _Win32Opener(filename, mode, fallback_mode)
      if _FcntlOpener:
        opener = _FcntlOpener(filename, mode, fallback_mode)

    if not opener:
      opener = _PosixOpener(filename, mode, fallback_mode)

    self._opener = opener

  def filename(self):
    """Return the filename we were constructed with."""
    return self._opener._filename

  def file_handle(self):
    """Return the file_handle to the opened file."""
    return self._opener.file_handle()

  def is_locked(self):
    """Return whether we successfully locked the file."""
    return self._opener.is_locked()

  def open_and_lock(self, timeout=0, delay=0.05):
    """Open the file, trying to lock it.

    Args:
      timeout: float, The number of seconds to try to acquire the lock.
      delay: float, The number of seconds to wait between retry attempts.

    Raises:
      AlreadyLockedException: if the lock is already acquired.
      IOError: if the open fails.
    """
    self._opener.open_and_lock(timeout, delay)

  def unlock_and_close(self):
    """Unlock and close a file."""
    self._opener.unlock_and_close()



from django.db import models
from rest import models as rm
from rest import helpers as rh
from location import geolocate


class GeoLocation(models.Model, rm.RestModel):
    city = models.CharField(max_length=64, blank=True, null=True)
    county = models.CharField(max_length=64)
    state = models.CharField(max_length=2)
    zip = models.CharField(max_length=32)
    country = models.CharField(max_length=64)
    area_codes = models.CharField(max_length=64)
    timezone = models.CharField(max_length=32)
    kind = models.CharField(max_length=32)

    lat = models.FloatField(default=0.0, blank=True)
    lng = models.FloatField(default=0.0, blank=True)

    @staticmethod
    def isValidZip(zipcode):
        return len(zipcode) > 2

    @staticmethod
    def findByZip(zipcode):
        return GeoLocation.objects.filter(zip=zipcode)

    @classmethod
    def byZip(cls, zipcode):
        model = cls.objects.filter(zip=zipcode).last()
        if model:
            return model
        try:
            loc = geolocate.search(zipcode).asDict()
            timezone = geolocate.getTimeZone(loc.get("latitude"), loc.get("longitude"))

            model = cls(
                city=loc.get("city"),
                county=loc.get("county", "n/a"),
                state=loc.get("short_administrative_area_level_1", "n/a"),
                zip=zipcode,
                country=loc.get("country"),
                timezone=timezone,
                lat=loc.get("latitude"),
                lng=loc.get("longitude"),
            )
            if model.county is None:
                model.county = "n/a"
            model.save()
        except Exception:
            rh.log_exception(f"byZip({zipcode})")
        return model

    @staticmethod
    def nearZip(zipcode, distance_miles=10):
        res = GeoLocation.objects.filter(zip=zipcode)
        # take the first one
        # TODO FIX ME
        distance = 0.01 * distance_miles
        if res.count():
            loc = res[0]
            # now adjust lat lng, poor mans approach
            blat = loc.lat + distance
            slat = loc.lat - distance
            blng = loc.lng + distance
            slng = loc.lng - distance
            res = GeoLocation.objects.filter(lat__lte=blat, lat__gte=slat, lng__lte=blng, lng__gte=slng)
        return res



from . import models as auditlog

from rest import views as rv
from rest import decorators as rd


@rd.url(r'^plog$')
@rd.url(r'^plog/(?P<plog_id>\d+)$')
@rd.perm_required('view_logs')
def plog_handler(request, plog_id=None):
    if not plog_id:
        min_pk = getattr(auditlog.settings, "PLOG_STALE_ID", 0)
        return auditlog.PersistentLog.on_rest_list(request, qset=auditlog.PersistentLog.objects.filter(pk__gte=min_pk))
    return auditlog.PersistentLog.on_rest_request(request, plog_id)


@rd.urlGET(r'^plog_old$')
@rd.perm_required('view_logs')
def plogList(request):
    auditlog.PersistentLog.on_request_handle()

    graph = request.DATA.get("graph", "default")
    qset = auditlog.PersistentLog.objects.all()
    if request.group:
        qset = qset.filter(group=request.group)

    ip = request.DATA.get("ip")
    if ip:
        qset = qset.filter(session__ip=ip)

    path = request.DATA.get("path")
    if path:
        qset = qset.filter(remote_path__icontains=path)
    
    method = request.DATA.get("method")
    if method:
        qset = qset.filter(method=method)

    action = request.DATA.get("action")
    if action:
        qset = qset.filter(action=action)
    
    component = request.DATA.get("component")
    if component:
        qset = qset.filter(component=component)
    
    pkey = request.DATA.get("pkey")
    if pkey:
        qset = qset.filter(pkey=pkey)

    username = request.DATA.get("username")
    if username:
        qset = qset.filter(user__username=username)
    
    term = request.DATA.get("term")
    if term:
        qset = qset.filter(message__icontains=term)
    
    return rv.restList(request, qset.order_by('-when'), **auditlog.PersistentLog.getGraph(graph))



from django.db import models as dm
from rest import models as rm
from medialib import models as medialib
import re
import mistune
from wiki.renderers import WikiRenderer
from wiki.renderers.mistune import task_list

class Page(dm.Model, rm.RestModel, rm.MetaDataModel):
    """
    Blog (a collection of articles)
    """
    class RestMeta:
        SEARCH_FIELDS = ["title", "body"]
        SEARCH_TERMS = ["title", "body"]
        QUERY_FIELDS = ["all_fields", "parent__path"]
        UNIQUE_LOOKUP = ["path"]
        DEFAULT_SORT = "-order"
        CAN_DELETE = True
        GRAPHS = {
            "basic": {
                "fields": [
                    "id",
                    "path",
                    "created",
                    "modified",
                    "title",
                    "parent",
                    "order",
                    "path",
                    "slug",
                ],
            },
            "default": {
                "fields": [
                    "id",
                    "path",
                    "created",
                    "modified",
                    "title",
                    "parent",
                    "order",
                    "slug",
                    "children_paths",
                    "body",
                    ("toHTML", "html")
                ],
                "recurse_into": ["media"],
                "graphs": {
                    "media": "default",
                    "member": "basic",
                    "parent": "basic"
                }
            },
            "rendered": {
                "graphs": {
                    "self": "default",
                    "member": "basic"
                }
            },
            "list": {
                "graphs": {
                    "self": "basic",
                    "member": "basic"
                }
            },
            "toc_child": {
                "graphs": {
                    "self": "basic",
                    "parent": "basic"
                }
            },
            "toc": {
                "extra": ["children"],
                "graphs": {
                    "self": "basic",
                    "children": "toc_child"
                }
            },
        }

    created = dm.DateTimeField(auto_now_add=True, editable=False)
    modified = dm.DateTimeField(db_index=True, auto_now=True)

    member = dm.ForeignKey(
        "account.Member", related_name="+",
        blank=True, null=True, default=None,
        on_delete=dm.SET_NULL)

    group = dm.ForeignKey(
        "account.Group", related_name="+", 
        blank=True, null=True, default=None,
        on_delete=dm.SET_NULL)

    parent = dm.ForeignKey(
        "Page", related_name="children",
        default=None, null=True, blank=True,
        on_delete=dm.CASCADE)

    order = dm.IntegerField(default=0, blank=True)

    title = dm.CharField(max_length=255)
    path = dm.CharField(max_length=255, db_index=True)
    slug = dm.SlugField(db_index=True)

    body = dm.TextField(blank=True)

    is_active = dm.BooleanField(blank=True, default=True)

    @property
    def children_paths(self):
        return list(self.children.all().values_list("path", flat=True))

    def upload__file(self, fobj, name):
        if fobj is None:
            return

        request = self.getActiveRequest()
        if not self.group:
            lib = request.member.getMediaLibrary("wiki")
        else:
            lib = self.group.getMediaLibrary("wiki")
        kind = medialib.MediaItem.guessMediaKind(fobj)
        media = medialib.MediaItem(library=lib, name=fobj.name, member=request.member, kind=kind, newfile=fobj)
        media.save()
        wmedia = PageMedia(media=media, entry=self, group=self.group, member=request.member)
        wmedia.save()

    def set_remove_media(self, value):
        wmedia = PageMedia.objects.filter(pk=value).last()
        if wmedia:
            wmedia.delete()
    
    def set_slug(self, value):
        sanitized_slug = value.strip().lower().replace(' ', '_')
        sanitized_slug = re.sub(r'\W+', '', sanitized_slug)
        sanitized_slug = re.sub(r'__+', '_', sanitized_slug).strip('_')
        if self.slug != sanitized_slug:
            self._recordRestChange("slug", self.slug)
            self.slug = sanitized_slug
            # confirm slug is unique
            qset = Page.objects.filter(parent=self.parent, slug=sanitized_slug)
            if qset.count():
                self.slug = f"{sanitized_slug}_{qset.count()+1}"

    def on_rest_pre_save(self, request):
        if not self.slug:
            self.set_slug(self.title)
        self.generatePath()

    def on_rest_saved(self, request, is_new=False):
        if self.hasFieldChanged("slug"):
            self.generatePath(True)

    def generatePath(self, propagate=False):
        paths = [self.slug]
        parent = self.parent
        while parent is not None:
            paths.append(parent.slug)
            parent = parent.parent
        paths.reverse()
        path = "/".join(paths)
        if self.path != path:
            self._recordRestChange("path", self.path)
            self.path = path
            if propagate:
                self.save()
        if propagate:
            for child in self.children.all():
                child.generatePath(propagate=True)
        return self.path

    def toHTML(self):
        renderer = WikiRenderer()
        md_engine = mistune.create_markdown(
            renderer=renderer,
            plugins=['def_list', 'url', 'abbr'])
        md_engine.use(task_list.plugin_task_lists)
        md = md_engine(self.body)
        return {
            "toc": renderer.render_toc(),
            "body": md
        }

    @classmethod
    def path_first_existing_parent(cls, path):
        paths = path.split("/")
        while len(paths):
            paths.pop()
            parent = cls.objects.filter(path="/".join(paths)).first()
            if parent is not None:
                return parent
        return None

    @classmethod
    def path_parent(cls, path):
        paths = path.split("/")
        paths.pop()
        return cls.objects.filter(path="/".join(paths)).first()


class PageMetaData(rm.MetaDataBase):
    parent = dm.ForeignKey(Page, related_name="properties", on_delete=dm.CASCADE)


class PageMedia(dm.Model, rm.RestModel):
    class RestMeta:
        GRAPHS = {
            "basic": {
                "graphs": {
                    "media": "default"
                }
            },
            "default": {
                "graphs": {
                    "media": "default"
                }
            }
        }
    created = dm.DateTimeField(auto_now_add=True, editable=False)

    member = dm.ForeignKey(
        "account.Member", related_name="+",
        blank=True, null=True, default=None,
        on_delete=dm.SET_NULL)

    group = dm.ForeignKey(
        "account.Group", related_name="+", 
        blank=True, null=True, default=None,
        on_delete=dm.SET_NULL)

    entry = dm.ForeignKey(Page, related_name="media_library", on_delete=dm.CASCADE)
    media = dm.ForeignKey("medialib.MediaItem", related_name="+", on_delete=dm.CASCADE)



from django.db import models
from datetime import datetime

from rest.models import RestModel
from rest import ua


# replacing legacy cookie session system with more robust session info
class AuthSession(models.Model, RestModel):
    class RestMeta:
        SEARCH_FIELDS = ["ip", "member__username", "browser", "location__city"]
        VIEW_PERMS = ["view_members", "manage_members", "manage_users", "owner"]
        CAN_SAVE = False
        CAN_DELETE = False
        VIEW_PERMS_MEMBER_FIELD = "member"
        # note the | will check collection parameter...
        #   trailing "." will check if the collection has the key set to true
        SEARCH_TERMS = [
            "ip", "device", "browser", "os",
            ("username", "member__username"),
            ("email", "member__email"),
            ("first_name", "member__first_name"),
            ("last_name", "member__last_name"),
            "last_activity#datetime"]
        GRAPHS = {
            "default": {
                "graphs": {
                    "location": "default"
                }
            },
            "location": {
                "graphs": {
                    "location": "default",
                    "member": "basic"
                }
            }
        }
    # the signature of session
    created = models.DateTimeField(auto_now_add=True, editable=False, db_index=True)
    last_activity = models.DateTimeField(db_index=True)

    signature = models.CharField(max_length=127, db_index=True)
    ip = models.CharField(max_length=127, null=True, blank=True, db_index=True)
    buid = models.CharField(max_length=127, null=True, blank=True, default=None)

    member = models.ForeignKey("account.Member", null=True, blank=True, related_name="auth_sessions", on_delete=models.CASCADE)
    location = models.ForeignKey("location.GeoIP", related_name="auth_sessions", blank=True, null=True, default=None, on_delete=models.CASCADE)

    user_agent = models.TextField(blank=True, null=True)
    browser = models.CharField(max_length=64, null=True, blank=True, default=None)
    os = models.CharField(max_length=64, null=True, blank=True, default=None)
    device = models.CharField(max_length=64, null=True, blank=True, default=None)

    def touch(self):
        self.last_activity = datetime.now()
        self.save()

    def updateLocation(self):
        if self.location is None and self.ip:
            GeoIP = self.getModel("location", "GeoIP")
            self.location = GeoIP.get(self.ip)
        return self.location

    def updateUA(self):
        if self.user_agent:
            ua_info = ua.parse(self.user_agent)
            self.browser = ua_info.get("user_agent.family")
            self.os = ua_info.get("os.family")
            self.device = ua_info.get("device.family")

    def __str__(self):
        return f"{self.member.username} - {self.ip} - {self.os} - {self.browser}"

    @classmethod
    def NewSession(cls, request):
        obj = cls(ip=request.ip, member=request.member, signature=request.signature)
        obj.user_agent = request.META.get('HTTP_USER_AGENT', None)
        obj.last_activity = datetime.now()
        obj.buid = request.POST.get("__buid__", request.GET.get("__buid__", None))
        obj.updateUA()
        obj.updateLocation()
        obj.save()
        return obj

    @classmethod
    def GetSession(cls, request, touch=True):
        if request.signature and request.member:
            session = AuthSession.objects.filter(signature=request.signature, member=request.member, ip=request.ip).last()
            if session is None:
                return cls.NewSession(request)
            if touch:
                session.touch()
            return session
        return None




from redis import ConnectionPool, StrictRedis

from ws4redis import settings
import time

from rest import UberDict
from rest.log import getLogger
logger = getLogger("async", filename="async.log")

REDIS_CON_POOL = None


def getRedisClient():
    global REDIS_CON_POOL
    if REDIS_CON_POOL is None:
        REDIS_CON_POOL = ConnectionPool(**settings.WS4REDIS_CONNECTION)
    return StrictRedis(connection_pool=REDIS_CON_POOL)


class RedisMessage(bytes):
    def __new__(cls, value):
        if isinstance(value, str):
            if value != settings.WS4REDIS_HEARTBEAT:
                return bytes(value, 'utf-8')
        elif isinstance(value, list):
            if len(value) >= 2 and value[0] == b'message':
                return value[2]
        elif isinstance(value, dict):
            if not hasattr(value, "toJSON"):
                value = UberDict(value)
            return bytes(value.toJSON(as_string=True), 'utf-8')
        elif isinstance(value, bytes):
            return value
        return None


class RedisStore():
    def __init__(self, connection=None):
        self.connection = connection
        if self.connection is None:
            self.connection = getRedisClient()
        self.subscriptions = []
        self.pubsub = None
        self.online_pk = None
        self.online_channel = None
        self.only_one = False  # don't count connections, only allows one instance
        self.expire = settings.WS4REDIS_EXPIRE

    def publish(self, message, channel, facility="events", pk=None, expire=None, prefix=settings.WS4REDIS_PREFIX):
        if expire is None:
            expire = self.expire
        if not isinstance(message, RedisMessage):
            message = RedisMessage(message)

        if not isinstance(message, bytes):
            raise ValueError('message is {} but should be bytes'.format(type(message)))

        if isinstance(pk, list):
            count = 0
            for spk in pk:
                count += self.publish(message, channel, facility, pk=spk, expire=expire, prefix=prefix)
            return count

        channel_key = self.channelToKey(channel, facility, pk, prefix)
        if settings.WS4REDIS_LOG_DEBUG:
            logger.info("publishing msg to: {0}".format(channel_key), message)
        count = self.connection.publish(channel_key, message)
        return count

    def getSubMessage(self):
        # get a message pending from subscription
        if self.pubsub:
            return self.pubsub.parse_response()
        return None

    def getPendingMessage(self, channel, facility="events", pk=None, prefix=settings.WS4REDIS_PREFIX):
        # get a message from the connection channel
        channel_key = self.channelToKey(channel, facility, pk, prefix)
        return self.connection.get(channel_key)

    def channelToKey(self, channel, facility="events", pk=None, prefix=settings.WS4REDIS_PREFIX):
        if not pk:
            key = F'{prefix}:{channel}:{facility}'
        else:
            key = F'{prefix}:{channel}:{pk}:{facility}'
        return key

    def subscribe(self, channel, facility="events", pk=None, prefix=settings.WS4REDIS_PREFIX):
        if self.pubsub is None:
            self.pubsub = self.connection.pubsub()
        key = self.channelToKey(channel, facility, pk, prefix)
        if key not in self.subscriptions:
            if settings.WS4REDIS_LOG_DEBUG:
                logger.info(F"subscribing to: {key}")
            self.subscriptions.append(key)
            self.pubsub.subscribe(key)
        return key

    def unsubscribe(self, channel, facility, pk=None, prefix=settings.WS4REDIS_PREFIX):
        key = self.channelToKey(channel, facility, pk, prefix)
        if key in self.subscriptions:
            if settings.WS4REDIS_LOG_DEBUG:
                logger.info(F"unsubscribing to: {key}")
            self.subscriptions.remove(key)
            self.pubsub.unsubscribe(key)

    def publishModelOnline(self, name, pk, only_one=False):
        if self.online_pk is None:
            self.online_pk = pk
            self.online_channel = name
            self.only_one = only_one
            if self.only_one:
                self.connection.sadd(F"{name}:online", pk)
            else:
                count = self.connection.hincrby(F"{name}:online:connections", pk, 1)
                if count == 1:
                    self.connection.sadd(F"{name}:online", pk)

    def unpublishModelOnline(self):
        if self.online_pk:
            name = self.online_channel
            pk = self.online_pk
            self.online_channel = None
            self.online_pk = None
            if self.only_one:
                self.connection.srem(F"{name}:online", pk)
            else:
                count = self.connection.hincrby(F"{name}:online:connections", pk, -1)
                if count == 0:
                    self.connection.srem(F"{name}:online", pk)

    def waitForMessage(self, muid=None, timeout=55):
        timeout_at = time.time() + timeout
        while time.time() < timeout_at:
            message = self.pubsub.get_message()
            if message is not None and message.get("type") == "message":
                return UberDict.fromJSON(message.get("data"))
            time.sleep(1.0)
        return None

    def get_file_descriptor(self):
        """
        Returns the file descriptor used for passing to the select call when listening
        on the message queue.
        """
        if self.pubsub.connection:
            return self.pubsub.connection._sock.fileno()
        return None

    def release(self):
        """
        New implementation to free up Redis subscriptions when websockets close. This prevents
        memory sap when Redis Output Buffer and Output Lists build when websockets are abandoned.
        """
        self.unpublishModelOnline()
        if self.pubsub and self.pubsub.subscribed:
            self.pubsub.unsubscribe()
            self.pubsub.reset()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.release()
        return False



from rest import helpers as rh


def cents_to_currency(cents, extra=None):
    currency, cents = divmod(cents, 100)
    return f"{currency}.{cents:02}"


def cents_to_dollars(cents, extra=None):
    dollars, cents = divmod(cents, 100)
    return f"${dollars}.{cents:02}"


def timezone(value, extra=None):
    value = rh.convertToLocalTime(extra, value, True)
    return value.strftime("%Y-%m-%d %H:%M:%S %Z")
